import re
import string
from abc import ABC, abstractmethod
from collections.abc import Iterable
from types import MappingProxyType
from typing import Any, Mapping, Protocol

from typing_extensions import assert_never

from phoenix.client.__generated__ import v1


class TemplateFormatter(Protocol):
    def format(
        self,
        template: str,
        /,
        *,
        variables: Mapping[str, str],
    ) -> str: ...


class BaseTemplateFormatter(ABC):
    @abstractmethod
    def parse(self, template: str) -> set[str]:
        """
        Parse the template and return a set of variable names.
        """
        raise NotImplementedError

    def format(
        self,
        template: str,
        /,
        *,
        variables: Mapping[str, str] = MappingProxyType({}),
    ) -> str:
        """
        Formats the template with the given variables.
        """
        template_variable_names = self.parse(template)
        if missing_template_variables := template_variable_names - set(variables.keys()):
            raise TemplateFormatterError(
                f"Missing template variable(s): {', '.join(missing_template_variables)}"
            )
        return self._format(template, template_variable_names, variables)

    @abstractmethod
    def _format(
        self,
        template: str,
        variable_names: Iterable[str],
        variables: Mapping[str, str] = MappingProxyType({}),
    ) -> str:
        raise NotImplementedError


class NoOpFormatterBase(BaseTemplateFormatter):
    """
    No-op template formatter.

    Examples:

    >>> formatter = NoOpFormatterBase()
    >>> formatter.format("hello")
    'hello'
    """

    def parse(self, template: str) -> set[str]:
        return set()

    def _format(self, template: str, *args: Any, **kwargs: Any) -> str:
        return template


class FStringBaseTemplateFormatter(BaseTemplateFormatter):
    """
    Regular f-string template formatter.

    Examples:

    >>> formatter = FStringBaseTemplateFormatter()
    >>> formatter.format("{hello}", {"hello": "world"})
    'world'
    """

    def parse(self, template: str) -> set[str]:
        return set(
            field_name for _, field_name, _, _ in string.Formatter().parse(template) if field_name
        )

    def _format(
        self,
        template: str,
        variable_names: Iterable[str],
        variables: Mapping[str, str] = MappingProxyType({}),
    ) -> str:
        return template.format(**variables)


class MustacheBaseTemplateFormatter(BaseTemplateFormatter):
    """
    Mustache template formatter.

    Examples:

    >>> formatter = MustacheBaseTemplateFormatter()
    >>> formatter.format("{{ hello }}", {"hello": "world"})
    'world'
    """

    PATTERN = re.compile(r"(?<!\\){{\s*(\w+)\s*}}")

    def parse(self, template: str) -> set[str]:
        return set(match for match in re.findall(self.PATTERN, template))

    def _format(
        self,
        template: str,
        variable_names: Iterable[str],
        variables: Mapping[str, str] = MappingProxyType({}),
    ) -> str:
        for variable_name in variable_names:
            pattern = rf"(?<!\\){{{{\s*{re.escape(variable_name)}\s*}}}}"
            replacement = variables[variable_name]
            template = re.sub(
                pattern=pattern,
                repl=lambda _: replacement,
                string=template,
            )
        return template


class TemplateFormatterError(Exception):
    """
    An error raised when template formatting fails.
    """

    pass


F_STRING_TEMPLATE_FORMATTER = FStringBaseTemplateFormatter()
MUSTACHE_TEMPLATE_FORMATTER = MustacheBaseTemplateFormatter()
NO_OP_FORMATTER = NoOpFormatterBase()


def to_formatter(obj: v1.PromptVersionData) -> BaseTemplateFormatter:
    if (
        "template_format" not in obj
        or not obj["template_format"]
        or obj["template_format"] == "MUSTACHE"
    ):
        return MUSTACHE_TEMPLATE_FORMATTER
    elif obj["template_format"] == "F_STRING":
        return F_STRING_TEMPLATE_FORMATTER
    elif obj["template_format"] == "NONE":
        return NO_OP_FORMATTER
    else:
        assert_never(obj["template_format"])



# pyright: reportPrivateUsage=false

import pandas as pd
import pytest

from phoenix.client.resources.annotations import _chunk_dataframe, _validate_dataframe


class TestDataFrameValidation:
    """Test suite for the _validate_dataframe helper function."""

    def test_empty(self) -> None:
        """Test validation of empty DataFrame."""
        df = pd.DataFrame()
        with pytest.raises(ValueError, match="DataFrame cannot be empty"):
            _validate_dataframe(dataframe=df)

    def test_not_pandas(self) -> None:
        """Test validation with non-pandas input."""
        with pytest.raises(TypeError, match="Expected pandas DataFrame"):
            _validate_dataframe(dataframe="not a dataframe")  # type: ignore[arg-type] # pyright: ignore[reportArgumentType]

    def test_missing_required_columns(self) -> None:
        """Test validation with missing required columns."""
        df = pd.DataFrame({"label": ["positive"]})
        with pytest.raises(
            ValueError, match="DataFrame must contain either 'name' or 'annotation_name' column"
        ):
            _validate_dataframe(dataframe=df)

    def test_both_name_columns(self) -> None:
        """Test validation when both name and annotation_name columns are present."""
        df = pd.DataFrame(
            {"name": ["sentiment"], "annotation_name": ["sentiment"], "annotator_kind": ["HUMAN"]}
        )
        with pytest.raises(
            ValueError, match="DataFrame cannot have both 'name' and 'annotation_name' columns"
        ):
            _validate_dataframe(dataframe=df)

    def test_invalid_annotator_kind(self) -> None:
        """Test validation of invalid annotator_kind values."""
        df = pd.DataFrame(
            {
                "name": ["sentiment"],
                "annotator_kind": ["INVALID"],
                "span_id": ["span1"],
            }
        )
        with pytest.raises(ValueError, match="Invalid annotator_kind values found in DataFrame"):
            _validate_dataframe(dataframe=df)

    def test_valid_with_name(self) -> None:
        """Test validation with valid DataFrame using name column."""
        df = pd.DataFrame(
            {"name": ["sentiment"], "annotator_kind": ["HUMAN"], "span_id": ["span1"]}
        )
        _validate_dataframe(dataframe=df)  # Should not raise

    def test_valid_with_annotation_name(self) -> None:
        """Test validation with valid DataFrame using annotation_name column."""
        df = pd.DataFrame(
            {"annotation_name": ["sentiment"], "annotator_kind": ["HUMAN"], "span_id": ["span1"]}
        )
        _validate_dataframe(dataframe=df)  # Should not raise

    def test_valid_with_global_name(self) -> None:
        """Test validation with valid DataFrame and global annotation_name."""
        df = pd.DataFrame({"annotator_kind": ["HUMAN"], "span_id": ["span1"]})
        _validate_dataframe(dataframe=df, annotation_name="sentiment")  # Should not raise

    def test_valid_with_global_annotator_kind(self) -> None:
        """Test validation with valid DataFrame and global annotator_kind."""
        df = pd.DataFrame({"name": ["sentiment"], "span_id": ["span1"]})
        _validate_dataframe(dataframe=df, annotator_kind="HUMAN")  # Should not raise

    def test_invalid_global_annotator_kind(self) -> None:
        """Test validation with invalid global annotator_kind."""
        df = pd.DataFrame({"name": ["sentiment"], "span_id": ["span1"]})
        with pytest.raises(ValueError, match="Invalid annotator_kind value"):
            _validate_dataframe(
                dataframe=df,
                annotator_kind="INVALID",  # type: ignore
            )

    def test_invalid_global_name(self) -> None:
        """Test validation with invalid global annotation_name."""
        df = pd.DataFrame({"annotator_kind": ["HUMAN"], "span_id": ["span1"]})
        with pytest.raises(ValueError, match="Annotation name cannot be empty or whitespace"):
            _validate_dataframe(dataframe=df, annotation_name="")  # Empty string

    def test_missing_span_id(self) -> None:
        """Test validation with missing span_id."""
        df = pd.DataFrame({"name": ["sentiment"], "annotator_kind": ["HUMAN"]})
        with pytest.raises(
            ValueError,
            match="DataFrame must have either a 'span_id' or 'context.span_id' column, or a string-based index",  # noqa: E501
        ):
            _validate_dataframe(dataframe=df)

    def test_both_span_id_columns(self) -> None:
        """Test validation when both span_id and context.span_id columns are present."""
        df = pd.DataFrame(
            {
                "name": ["sentiment"],
                "annotator_kind": ["HUMAN"],
                "span_id": ["span1"],
                "context.span_id": ["span1"],
            }
        )
        with pytest.raises(
            ValueError, match="DataFrame cannot have both 'span_id' and 'context.span_id' columns"
        ):
            _validate_dataframe(dataframe=df)

    def test_valid_with_context_span_id(self) -> None:
        """Test validation with valid DataFrame using context.span_id column."""
        df = pd.DataFrame(
            {
                "name": ["sentiment"],
                "annotator_kind": ["HUMAN"],
                "context.span_id": ["span1"],
            }
        )
        _validate_dataframe(dataframe=df)  # Should not raise

    def test_invalid_context_span_id_values(self) -> None:
        """Test validation with invalid context.span_id values."""
        df = pd.DataFrame(
            {
                "name": ["sentiment", "sentiment"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "context.span_id": ["", "  "],  # Empty strings
            }
        )
        with pytest.raises(ValueError, match="context.span_id values must be non-empty strings"):
            _validate_dataframe(dataframe=df)

    def test_none_context_span_id_values(self) -> None:
        """Test validation with None values in context.span_id column."""
        df = pd.DataFrame(
            {
                "name": ["sentiment", "sentiment"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "context.span_id": [None, "valid_id"],  # None value
            }
        )
        with pytest.raises(ValueError, match="context.span_id values cannot be None"):
            _validate_dataframe(dataframe=df)

    def test_non_string_context_span_id(self) -> None:
        """Test validation with non-string values in context.span_id column."""
        df = pd.DataFrame(
            {
                "name": ["sentiment", "sentiment"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "context.span_id": [123, "valid_id"],  # Non-string value
            }
        )
        with pytest.raises(ValueError, match="context.span_id values must be strings"):
            _validate_dataframe(dataframe=df)

    def test_valid_with_index(self) -> None:
        """Test validation with valid DataFrame using index as span_id."""
        df = pd.DataFrame({"name": ["sentiment"], "annotator_kind": ["HUMAN"]}, index=["span1"])
        _validate_dataframe(dataframe=df)  # Should not raise

    def test_invalid_name_values(self) -> None:
        """Test validation with invalid name values."""
        df = pd.DataFrame(
            {
                "name": ["", "  "],  # Empty strings
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["span1", "span2"],
            }
        )
        with pytest.raises(ValueError, match="name values must be non-empty strings"):
            _validate_dataframe(dataframe=df)

    def test_invalid_annotation_name_values(self) -> None:
        """Test validation with invalid annotation_name values."""
        df = pd.DataFrame(
            {
                "annotation_name": ["", "  "],  # Empty strings
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["span1", "span2"],
            }
        )
        with pytest.raises(ValueError, match="annotation_name values must be non-empty strings"):
            _validate_dataframe(dataframe=df)

    def test_invalid_span_id_values(self) -> None:
        """Test validation with invalid span_id values."""
        df = pd.DataFrame(
            {
                "name": ["sentiment", "sentiment"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["", "  "],  # Empty strings
            }
        )
        with pytest.raises(ValueError, match="span_id values must be non-empty strings"):
            _validate_dataframe(dataframe=df)

    def test_invalid_index_values(self) -> None:
        """Test validation with invalid index values when using index as span_id."""
        df = pd.DataFrame(
            {
                "name": ["sentiment", "sentiment"],
                "annotator_kind": ["HUMAN", "HUMAN"],
            },
            index=["", "  "],  # Empty strings
        )
        with pytest.raises(
            ValueError, match="Index values must be non-empty strings when used as span_id"
        ):
            _validate_dataframe(dataframe=df)

    def test_mixed_valid_invalid_values(self) -> None:
        """Test validation with mixed valid and invalid values in columns."""
        df = pd.DataFrame(
            {
                "name": ["valid", ""],  # Mixed valid/invalid
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["valid_id", "  "],  # Mixed valid/invalid
            }
        )
        with pytest.raises(ValueError, match="name values must be non-empty strings"):
            _validate_dataframe(dataframe=df)

    def test_non_string_span_id(self) -> None:
        """Test validation with non-string values in span_id column."""
        df = pd.DataFrame(
            {
                "name": ["sentiment", "sentiment"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": [123, "valid_id"],  # Non-string value
            }
        )
        with pytest.raises(ValueError, match="span_id values must be strings"):
            _validate_dataframe(dataframe=df)

    def test_whitespace_only_index(self) -> None:
        """Test validation with whitespace-only values in index."""
        df = pd.DataFrame(
            {
                "name": ["sentiment", "sentiment"],
                "annotator_kind": ["HUMAN", "HUMAN"],
            },
            index=["  ", "\t"],  # Whitespace-only values
        )
        with pytest.raises(
            ValueError, match="Index values must be non-empty strings when used as span_id"
        ):
            _validate_dataframe(dataframe=df)


class TestChunkDataFrame:
    """Test suite for the _chunk_dataframe helper function."""

    def test_empty_dataframe(self) -> None:
        """Test chunking an empty DataFrame."""
        df = pd.DataFrame()
        with pytest.raises(ValueError, match="DataFrame cannot be empty"):
            list(_chunk_dataframe(dataframe=df))

    def test_single_row(self) -> None:
        """Test chunking a DataFrame with a single row."""
        df = pd.DataFrame(
            {
                "name": ["test1"],
                "annotator_kind": ["HUMAN"],
                "span_id": ["id1"],
                "label": ["label1"],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df))
        assert len(chunks) == 1
        assert len(chunks[0]) == 1

    def test_default_chunk_size(self) -> None:
        """Test that the default chunk size is used when not specified."""
        # Create a DataFrame with 101 rows to test default chunk size of 100
        df = pd.DataFrame(
            {
                "name": [f"test{i}" for i in range(101)],
                "annotator_kind": ["HUMAN"] * 101,
                "span_id": [f"id{i}" for i in range(101)],
                "label": [f"label{i}" for i in range(101)],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df))
        assert len(chunks) == 2
        assert len(chunks[0]) == 100  # First chunk should be full
        assert len(chunks[1]) == 1  # Second chunk should have remaining row

    def test_exact_chunk_size(self) -> None:
        """Test chunking a DataFrame that is exactly the chunk size."""
        # Create a DataFrame with 100 rows (exactly the default chunk size)
        df = pd.DataFrame(
            {
                "name": [f"test{i}" for i in range(100)],
                "annotator_kind": ["HUMAN"] * 100,
                "span_id": [f"id{i}" for i in range(100)],
                "label": [f"label{i}" for i in range(100)],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df))
        assert len(chunks) == 1
        assert len(chunks[0]) == 100

    def test_custom_chunk_size(self) -> None:
        """Test chunking with a custom chunk size."""
        # Create a DataFrame with 10 rows and use chunk size of 3
        df = pd.DataFrame(
            {
                "name": [f"test{i}" for i in range(10)],
                "annotator_kind": ["HUMAN"] * 10,
                "span_id": [f"id{i}" for i in range(10)],
                "label": [f"label{i}" for i in range(10)],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df, chunk_size=3))
        assert len(chunks) == 4  # 3 full chunks + 1 partial chunk
        assert len(chunks[0]) == 3
        assert len(chunks[1]) == 3
        assert len(chunks[2]) == 3
        assert len(chunks[3]) == 1

    def test_global_annotation_name(self) -> None:
        """Test chunking with global annotation_name."""
        df = pd.DataFrame(
            {
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["id1", "id2"],
                "label": ["label1", "label2"],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df, annotation_name="global_name"))
        assert len(chunks) == 1
        assert all(anno["name"] == "global_name" for anno in chunks[0])

    def test_global_annotator_kind(self) -> None:
        """Test chunking with global annotator_kind."""
        df = pd.DataFrame(
            {
                "name": ["test1", "test2"],
                "span_id": ["id1", "id2"],
                "label": ["label1", "label2"],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df, annotator_kind="HUMAN"))
        assert len(chunks) == 1
        assert all(anno["annotator_kind"] == "HUMAN" for anno in chunks[0])

    def test_optional_fields(self) -> None:
        """Test chunking with optional fields (label, score, explanation)."""
        df = pd.DataFrame(
            {
                "name": ["test1", "test2"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["id1", "id2"],
                "label": ["label1", "label2"],
                "score": [0.5, 0.8],
                "explanation": ["expl1", "expl2"],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df))
        assert len(chunks) == 1
        assert chunks[0][0]["result"]["label"] == "label1"  # pyright: ignore[reportTypedDictNotRequiredAccess]
        assert chunks[0][0]["result"]["score"] == 0.5  # pyright: ignore[reportTypedDictNotRequiredAccess]
        assert chunks[0][0]["result"]["explanation"] == "expl1"  # pyright: ignore[reportTypedDictNotRequiredAccess]

    def test_index_as_span_id(self) -> None:
        """Test chunking when using index as span_id."""
        df = pd.DataFrame(
            {
                "name": ["test1", "test2"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "label": ["label1", "label2"],
            },
            index=["id1", "id2"],
        )
        chunks = list(_chunk_dataframe(dataframe=df))
        assert len(chunks) == 1
        assert chunks[0][0]["span_id"] == "id1"
        assert chunks[0][1]["span_id"] == "id2"

    def test_metadata(self) -> None:
        """Test chunking with metadata field."""
        df = pd.DataFrame(
            {
                "name": ["test1", "test2"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["id1", "id2"],
                "label": ["label1", "label2"],
                "metadata": [{"key1": "value1"}, {"key2": "value2"}],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df))
        assert len(chunks) == 1
        assert chunks[0][0]["metadata"] == {"key1": "value1"}  # pyright: ignore[reportTypedDictNotRequiredAccess]
        assert chunks[0][1]["metadata"] == {"key2": "value2"}  # pyright: ignore[reportTypedDictNotRequiredAccess]

    def test_identifier(self) -> None:
        """Test chunking with identifier field."""
        df = pd.DataFrame(
            {
                "name": ["test1", "test2"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["id1", "id2"],
                "label": ["label1", "label2"],
                "identifier": ["id1", "id2"],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df))
        assert len(chunks) == 1
        assert chunks[0][0]["identifier"] == "id1"  # pyright: ignore[reportTypedDictNotRequiredAccess]
        assert chunks[0][1]["identifier"] == "id2"  # pyright: ignore[reportTypedDictNotRequiredAccess]

    def test_invalid_score_type(self) -> None:
        """Test chunking with invalid score type."""
        df = pd.DataFrame(
            {
                "name": ["test1"],
                "annotator_kind": ["HUMAN"],
                "span_id": ["id1"],
                "label": ["label1"],
                "score": ["not_a_number"],
            }
        )
        with pytest.raises(
            ValueError,
            match="Error processing row 0: Score value 'not_a_number' cannot be converted to float",
        ):
            list(_chunk_dataframe(dataframe=df))

    def test_chunk_with_context_span_id(self) -> None:
        """Test chunking with context.span_id column."""
        df = pd.DataFrame(
            {
                "name": ["test1", "test2"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "context.span_id": ["id1", "id2"],
                "label": ["label1", "label2"],
            }
        )
        chunks = list(_chunk_dataframe(dataframe=df))
        assert len(chunks) == 1
        assert chunks[0][0]["span_id"] == "id1"
        assert chunks[0][1]["span_id"] == "id2"

    def test_chunk_with_both_span_id_columns(self) -> None:
        """Test chunking with both span_id and context.span_id columns."""
        df = pd.DataFrame(
            {
                "name": ["test1", "test2"],
                "annotator_kind": ["HUMAN", "HUMAN"],
                "span_id": ["id1", "id2"],
                "context.span_id": ["id1", "id2"],
                "label": ["label1", "label2"],
            }
        )
        with pytest.raises(
            ValueError, match="DataFrame cannot have both 'span_id' and 'context.span_id' columns"
        ):
            list(_chunk_dataframe(dataframe=df))



from typing import Any, Dict

import pytest

from phoenix.client.types.spans import (
    Concatenation,
    Explosion,
    Projection,
    SpanFilter,
    SpanQuery,
)
from phoenix.trace.dsl import SpanQuery as PhoenixSpanQuery


def assert_dict_equivalence(dict1: Dict[str, Any], dict2: Dict[str, Any]) -> None:
    assert dict1.keys() == dict2.keys()
    for key in dict1:
        if isinstance(dict1[key], dict):
            assert_dict_equivalence(dict1[key], dict2[key])
        else:
            assert dict1[key] == dict2[key]


def test_basic_query_equivalence() -> None:
    phoenix_query = PhoenixSpanQuery().select("span_id", "trace_id")
    client_query = SpanQuery().select("span_id", "trace_id")

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())


def test_filter_query_equivalence() -> None:
    phoenix_query = PhoenixSpanQuery().where("span_id == '123'")
    client_query = SpanQuery().where("span_id == '123'")

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())


def test_explode_query_equivalence() -> None:
    phoenix_query = PhoenixSpanQuery().explode("attributes")
    client_query = SpanQuery().explode("attributes")

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())


def test_concat_query_equivalence() -> None:
    phoenix_query = PhoenixSpanQuery().concat("messages")
    client_query = SpanQuery().concat("messages")

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())


def test_rename_query_equivalence() -> None:
    phoenix_query = PhoenixSpanQuery().rename(old_name="new_name")
    client_query = SpanQuery().rename(old_name="new_name")

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())


def test_index_query_equivalence() -> None:
    phoenix_query = PhoenixSpanQuery().with_index("span_id")
    client_query = SpanQuery().with_index("span_id")

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())


def test_complex_query_equivalence() -> None:
    phoenix_query = (
        PhoenixSpanQuery()
        .select("span_id", "trace_id")
        .where("span_id == '123'")
        .explode("attributes")
        .concat("messages")
        .rename(old_name="new_name")
        .with_index("span_id")
    )
    client_query = (
        SpanQuery()
        .select("span_id", "trace_id")
        .where("span_id == '123'")
        .explode("attributes")
        .concat("messages")
        .rename(old_name="new_name")
        .with_index("span_id")
    )

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())


def test_backward_compatibility() -> None:
    query = SpanQuery().select(
        "context.span_id",
        "context.trace_id",
        "cumulative_token_count.completion",
    )

    # The query should internally convert to new field names
    query_dict = query.to_dict()
    assert "select" in query_dict
    select_dict = query_dict["select"]
    assert "span_id" in select_dict
    assert "trace_id" in select_dict
    assert "cumulative_llm_token_count_completion" in select_dict


def test_empty_key_validation() -> None:
    with pytest.raises(ValueError):
        Projection(key="")

    with pytest.raises(ValueError):
        SpanFilter(condition="")

    with pytest.raises(ValueError):
        Explosion(key="")

    with pytest.raises(ValueError):
        Concatenation(key="")


def test_chaining_order_equivalence() -> None:
    q1 = SpanQuery().select("span_id", "trace_id").where("span_id == '123'")
    q2 = SpanQuery().where("span_id == '123'").select("span_id", "trace_id")
    assert_dict_equivalence(q1.to_dict(), q2.to_dict())


def test_default_index_behavior() -> None:
    query = SpanQuery().select("span_id")
    phoenix_query = PhoenixSpanQuery().select("span_id")
    assert_dict_equivalence(query.to_dict(), phoenix_query.to_dict())


def test_explosion_kwargs_equivalence() -> None:
    phoenix_query = PhoenixSpanQuery().explode("attributes", arg1="span_id", arg2="some_field")
    client_query = SpanQuery().explode("attributes", arg1="span_id", arg2="some_field")

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())


def test_concatenation_kwargs_equivalence() -> None:
    phoenix_query = PhoenixSpanQuery().concat("messages", arg1="span_id", arg2="some_field")
    client_query = SpanQuery().concat("messages", arg1="span_id", arg2="some_field")

    assert_dict_equivalence(phoenix_query.to_dict(), client_query.to_dict())



from __future__ import annotations

import importlib.metadata
import json
import logging
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    Optional,
    Sequence,
    TypedDict,
    Union,
    cast,
    overload,
)

from typing_extensions import Required, TypeAlias, assert_never

from phoenix.client.__generated__ import v1
from phoenix.client.utils.template_formatters import TemplateFormatter, to_formatter

try:
    _anthropic_version = cast(
        tuple[int, int], tuple(map(int, importlib.metadata.version("anthropic").split(".")[:2]))
    )
except importlib.metadata.PackageNotFoundError:
    _anthropic_version = (0, 0)

if TYPE_CHECKING:
    from anthropic._client import Anthropic
    from anthropic.types import (
        ContentBlock,
        DocumentBlockParam,
        ImageBlockParam,
        MessageParam,
        ModelParam,
        RedactedThinkingBlockParam,
        TextBlock,
        TextBlockParam,
        ThinkingBlockParam,
        ThinkingConfigDisabledParam,
        ThinkingConfigEnabledParam,
        ToolChoiceAnyParam,
        ToolChoiceAutoParam,
        ToolChoiceParam,
        ToolChoiceToolParam,
        ToolParam,
        ToolResultBlockParam,
        ToolUnionParam,
        ToolUseBlock,
        ToolUseBlockParam,
    )
    from anthropic.types.message_create_params import MessageCreateParamsBase

    _BlockParam: TypeAlias = Union[
        TextBlockParam,
        ImageBlockParam,
        ToolUseBlockParam,
        ToolResultBlockParam,
        ThinkingBlockParam,
        RedactedThinkingBlockParam,
        DocumentBlockParam,
    ]
    _ContentPart: TypeAlias = Union[
        v1.TextContentPart,
        v1.ToolCallContentPart,
        v1.ToolResultContentPart,
    ]

    def _(obj: v1.PromptVersionData) -> None:
        messages, kwargs = to_chat_messages_and_kwargs(obj)
        Anthropic().messages.create(messages=messages, **kwargs)


class _ToolKwargs(TypedDict, total=False):
    tools: list[ToolParam]
    tool_choice: ToolChoiceParam


class _InvocationParameters(TypedDict, total=False):
    max_tokens: Required[int]
    stop_sequences: list[str]
    temperature: float
    top_p: float
    thinking: Union[ThinkingConfigEnabledParam, ThinkingConfigDisabledParam]


class AnthropicMessageModelKwargs(
    _InvocationParameters,
    _ToolKwargs,
    TypedDict,
    total=False,
):
    model: Required[ModelParam]
    system: Union[str, list[TextBlockParam]]


logger = logging.getLogger(__name__)


__all__ = [
    "create_prompt_version_from_anthropic",
    "to_chat_messages_and_kwargs",
]


def create_prompt_version_from_anthropic(
    obj: MessageCreateParamsBase,
    /,
    *,
    description: Optional[str] = None,
    template_format: Literal["F_STRING", "MUSTACHE", "NONE"] = "MUSTACHE",
    model_provider: Literal["ANTHROPIC"] = "ANTHROPIC",
) -> v1.PromptVersionData:
    invocation_parameters = _InvocationParametersConversion.from_anthropic(
        obj,
        model_provider=model_provider,
    )
    messages: list[v1.PromptMessage] = []
    if "system" in obj:
        system = (
            obj["system"]
            if isinstance(obj["system"], str)
            else list(map(_TextContentPartConversion.from_anthropic, obj["system"]))
        )
        messages.append(v1.PromptMessage(role="system", content=system))
    messages.extend(map(_MessageConversion.from_anthropic, obj["messages"]))
    template: v1.PromptChatTemplate = {
        "type": "chat",
        "messages": messages,
    }
    ans = v1.PromptVersionData(
        model_provider=model_provider,
        model_name=obj["model"],
        template=template,
        template_type="CHAT",
        template_format=template_format,
        invocation_parameters=invocation_parameters,
    )
    tool_kwargs: _ToolKwargs = {}
    if "tools" in obj:
        tool_kwargs["tools"] = list(_get_tool_params(obj["tools"]))
    if "tool_choice" in obj:
        tool_kwargs["tool_choice"] = obj["tool_choice"]
    if (tools := _ToolKwargsConversion.from_anthropic(tool_kwargs)) is not None:
        ans["tools"] = tools
    if description:
        ans["description"] = description
    return ans


def to_chat_messages_and_kwargs(
    obj: v1.PromptVersionData,
    /,
    *,
    variables: Mapping[str, str] = MappingProxyType({}),
    formatter: Optional[TemplateFormatter] = None,
) -> tuple[list[MessageParam], AnthropicMessageModelKwargs]:
    formatter = formatter or to_formatter(obj)
    assert formatter is not None
    template = obj["template"]
    system_messages: list[str] = []
    messages: list[MessageParam] = []
    if template["type"] == "chat":
        for message in template["messages"]:
            if message["role"] == "system":
                if isinstance(message["content"], str):
                    system_messages.append(message["content"])
                    continue
                for block in _ContentConversion.to_anthropic(
                    message["content"], variables, formatter
                ):
                    if block["type"] == "text":
                        system_messages.append(block["text"])
            else:
                messages.extend(_MessageConversion.to_anthropic(message, variables, formatter))
    elif template["type"] == "string":
        raise NotImplementedError
    else:
        assert_never(template)
    kwargs: AnthropicMessageModelKwargs = _ModelKwargsConversion.to_anthropic(obj)
    if system_messages:
        if len(system_messages) == 1:
            kwargs["system"] = system_messages[0]
        else:
            kwargs["system"] = [{"type": "text", "text": text} for text in system_messages]
    return messages, kwargs


_DEFAULT_MAX_TOKENS = 100


class _ModelKwargsConversion:
    @staticmethod
    def to_anthropic(
        obj: v1.PromptVersionData,
    ) -> AnthropicMessageModelKwargs:
        parameters: _InvocationParameters = (
            _InvocationParametersConversion.to_anthropic(obj["invocation_parameters"])
            if "invocation_parameters" in obj
            else _InvocationParameters(max_tokens=_DEFAULT_MAX_TOKENS)
        )
        ans: AnthropicMessageModelKwargs = {
            "model": obj["model_name"],
            **parameters,  # type: ignore[typeddict-item]
        }
        if "tools" in obj:
            tool_kwargs = _ToolKwargsConversion.to_anthropic(obj["tools"])
            if "tools" in tool_kwargs:
                ans["tools"] = tool_kwargs["tools"]
                if "tool_choice" in tool_kwargs:
                    ans["tool_choice"] = tool_kwargs["tool_choice"]
        return ans


class _InvocationParametersConversion:
    @staticmethod
    def to_anthropic(
        obj: Union[
            v1.PromptOpenAIInvocationParameters,
            v1.PromptAzureOpenAIInvocationParameters,
            v1.PromptAnthropicInvocationParameters,
            v1.PromptGoogleInvocationParameters,
            v1.PromptAwsInvocationParameters,
        ],
    ) -> _InvocationParameters:
        ans: _InvocationParameters = _InvocationParameters(
            max_tokens=_DEFAULT_MAX_TOKENS,
        )
        if obj["type"] == "anthropic":
            anthropic_params: v1.PromptAnthropicInvocationParametersContent
            anthropic_params = obj["anthropic"]
            if "max_tokens" in anthropic_params:
                ans["max_tokens"] = anthropic_params["max_tokens"]
            if "temperature" in anthropic_params:
                ans["temperature"] = anthropic_params["temperature"]
            if "top_p" in anthropic_params:
                ans["top_p"] = anthropic_params["top_p"]
            if "stop_sequences" in anthropic_params:
                ans["stop_sequences"] = list(anthropic_params["stop_sequences"])
            if "thinking" in anthropic_params:
                thinking = anthropic_params["thinking"]
                if thinking["type"] == "disabled":
                    ans["thinking"] = {
                        "type": "disabled",
                    }
                elif thinking["type"] == "enabled":
                    ans["thinking"] = {
                        "type": "enabled",
                        "budget_tokens": thinking["budget_tokens"],
                    }
                elif TYPE_CHECKING:
                    assert_never(thinking["type"])
        elif obj["type"] == "openai":
            openai_params: v1.PromptOpenAIInvocationParametersContent
            openai_params = obj["openai"]
            if "max_tokens" in openai_params:
                ans["max_tokens"] = openai_params["max_tokens"]
            if "temperature" in openai_params:
                ans["temperature"] = openai_params["temperature"]
            if "top_p" in openai_params:
                ans["top_p"] = openai_params["top_p"]
        elif obj["type"] == "azure_openai":
            azure_params: v1.PromptAzureOpenAIInvocationParametersContent
            azure_params = obj["azure_openai"]
            if "max_tokens" in azure_params:
                ans["max_tokens"] = azure_params["max_tokens"]
            if "temperature" in azure_params:
                ans["temperature"] = azure_params["temperature"]
            if "top_p" in azure_params:
                ans["top_p"] = azure_params["top_p"]
        elif obj["type"] == "google":
            google_params: v1.PromptGoogleInvocationParametersContent
            google_params = obj["google"]
            if "max_output_tokens" in google_params:
                ans["max_tokens"] = google_params["max_output_tokens"]
            if "temperature" in google_params:
                ans["temperature"] = google_params["temperature"]
            if "top_p" in google_params:
                ans["top_p"] = google_params["top_p"]
        elif obj["type"] == "aws":
            aws_params: v1.PromptAwsInvocationParametersContent
            aws_params = obj["aws"]
            if "max_tokens" in aws_params:
                ans["max_tokens"] = aws_params["max_tokens"]
            if "temperature" in aws_params:
                ans["temperature"] = aws_params["temperature"]
            if "top_p" in aws_params:
                ans["top_p"] = aws_params["top_p"]
        elif TYPE_CHECKING:
            assert_never(obj["type"])
        return ans

    @staticmethod
    def from_anthropic(
        obj: MessageCreateParamsBase,
        /,
        *,
        model_provider: Literal["ANTHROPIC"] = "ANTHROPIC",
    ) -> v1.PromptAnthropicInvocationParameters:
        content = v1.PromptAnthropicInvocationParametersContent(
            max_tokens=obj["max_tokens"],
        )
        if "temperature" in obj:
            content["temperature"] = obj["temperature"]
        if "top_p" in obj:
            content["top_p"] = obj["top_p"]
        if "stop_sequences" in obj:
            content["stop_sequences"] = list(obj["stop_sequences"])
        if "thinking" in obj:
            thinking = obj["thinking"]
            if thinking["type"] == "disabled":
                content["thinking"] = {
                    "type": "disabled",
                }
            elif thinking["type"] == "enabled":
                content["thinking"] = {
                    "type": "enabled",
                    "budget_tokens": thinking["budget_tokens"],
                }
            elif TYPE_CHECKING:
                assert_never(thinking["type"])
        return v1.PromptAnthropicInvocationParameters(
            type="anthropic",
            anthropic=content,
        )


class _ToolKwargsConversion:
    @staticmethod
    def to_anthropic(
        obj: Optional[v1.PromptTools],
    ) -> _ToolKwargs:
        ans: _ToolKwargs = {}
        if not obj:
            return ans
        tools: list[ToolParam] = list(_ToolConversion.to_anthropic(obj["tools"]))
        ans["tools"] = tools
        if "tool_choice" in obj:
            if obj["tool_choice"]["type"] == "none":
                return {}
            disable_parallel_tool_use: Optional[bool] = (
                obj["disable_parallel_tool_calls"] if "disable_parallel_tool_calls" in obj else None
            )
            tool_choice: ToolChoiceParam = _ToolChoiceConversion.to_anthropic(
                obj["tool_choice"],
                disable_parallel_tool_use,
            )
            ans["tool_choice"] = tool_choice
        return ans

    @staticmethod
    def from_anthropic(
        obj: _ToolKwargs,
    ) -> Optional[v1.PromptTools]:
        if not obj or "tools" not in obj:
            return None
        tools: list[v1.PromptToolFunction] = list(_ToolConversion.from_anthropic(obj["tools"]))
        if not tools:
            return None
        ans = v1.PromptTools(
            type="tools",
            tools=tools,
        )
        if "tool_choice" in obj:
            tc: ToolChoiceParam = obj["tool_choice"]
            tool_choice, disable_parallel_tool_use = _ToolChoiceConversion.from_anthropic(tc)
            ans["tool_choice"] = tool_choice
            if disable_parallel_tool_use is not None:
                ans["disable_parallel_tool_calls"] = disable_parallel_tool_use
        return ans


class _ToolChoiceConversion:
    @staticmethod
    def to_anthropic(
        obj: Union[
            v1.PromptToolChoiceNone,
            v1.PromptToolChoiceZeroOrMore,
            v1.PromptToolChoiceOneOrMore,
            v1.PromptToolChoiceSpecificFunctionTool,
        ],
        disable_parallel_tool_use: Optional[bool] = None,
    ) -> ToolChoiceParam:
        if obj["type"] == "none":
            return {"type": "none"}
        if obj["type"] == "zero_or_more":
            choice_auto: ToolChoiceAutoParam = {"type": "auto"}
            if disable_parallel_tool_use is not None:
                choice_auto["disable_parallel_tool_use"] = disable_parallel_tool_use
            return choice_auto
        if obj["type"] == "one_or_more":
            choice_any: ToolChoiceAnyParam = {"type": "any"}
            if disable_parallel_tool_use is not None:
                choice_any["disable_parallel_tool_use"] = disable_parallel_tool_use
            return choice_any
        if obj["type"] == "specific_function":
            choice_tool: ToolChoiceToolParam = {"type": "tool", "name": obj["function_name"]}
            if disable_parallel_tool_use is not None:
                choice_tool["disable_parallel_tool_use"] = disable_parallel_tool_use
            return choice_tool
        assert_never(obj["type"])

    @staticmethod
    def from_anthropic(
        obj: ToolChoiceParam,
    ) -> tuple[
        Union[
            v1.PromptToolChoiceNone,
            v1.PromptToolChoiceZeroOrMore,
            v1.PromptToolChoiceOneOrMore,
            v1.PromptToolChoiceSpecificFunctionTool,
        ],
        Optional[bool],
    ]:
        if obj["type"] == "auto":
            disable_parallel_tool_use = (
                obj["disable_parallel_tool_use"] if "disable_parallel_tool_use" in obj else None
            )
            choice_zero_or_more: v1.PromptToolChoiceZeroOrMore = {"type": "zero_or_more"}
            return choice_zero_or_more, disable_parallel_tool_use
        if obj["type"] == "any":
            disable_parallel_tool_use = (
                obj["disable_parallel_tool_use"] if "disable_parallel_tool_use" in obj else None
            )
            choice_one_or_more: v1.PromptToolChoiceOneOrMore = {"type": "one_or_more"}
            return choice_one_or_more, disable_parallel_tool_use
        if obj["type"] == "tool":
            disable_parallel_tool_use = (
                obj["disable_parallel_tool_use"] if "disable_parallel_tool_use" in obj else None
            )
            choice_function_tool: v1.PromptToolChoiceSpecificFunctionTool = {
                "type": "specific_function",
                "function_name": obj["name"],
            }
            return choice_function_tool, disable_parallel_tool_use
        if obj["type"] == "none":
            return v1.PromptToolChoiceNone(type="none"), None
        assert_never(obj["type"])


class _ToolConversion:
    @staticmethod
    def to_anthropic(
        obj: Iterable[v1.PromptToolFunction],
    ) -> Iterator[ToolParam]:
        for tool in obj:
            function = tool["function"]
            input_schema: dict[str, Any] = (
                dict(function["parameters"]) if "parameters" in function else {}
            )
            param: ToolParam = {
                "name": function["name"],
                "input_schema": input_schema,
            }
            if "description" in function:
                param["description"] = function["description"]
            yield param

    @staticmethod
    def from_anthropic(
        obj: Iterable[ToolParam],
    ) -> Iterator[v1.PromptToolFunction]:
        for tool in obj:
            function = v1.PromptToolFunctionDefinition(
                name=tool["name"],
            )
            if "description" in tool:
                function["description"] = tool["description"]
            if "input_schema" in tool:
                function["parameters"] = tool["input_schema"]
            yield v1.PromptToolFunction(type="function", function=function)


class _MessageConversion:
    @staticmethod
    def to_anthropic(
        obj: v1.PromptMessage,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> Iterator[MessageParam]:
        role = _RoleConversion.to_anthropic(obj)
        if isinstance(obj["content"], str):
            yield {"role": role, "content": formatter.format(obj["content"], variables=variables)}
            return
        blocks = list(_ContentConversion.to_anthropic(obj["content"], variables, formatter))
        if len(blocks) == 1 and blocks[0]["type"] == "text":
            yield {"role": role, "content": blocks[0]["text"]}
            return
        yield {"role": role, "content": blocks}

    @staticmethod
    def from_anthropic(
        obj: MessageParam,
    ) -> v1.PromptMessage:
        content = _ContentConversion.from_anthropic(obj["content"])
        role = _RoleConversion.from_anthropic(obj)
        return v1.PromptMessage(role=role, content=content)


class _TextContentPartConversion:
    @staticmethod
    def to_anthropic(
        obj: v1.TextContentPart,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> TextBlockParam:
        text = formatter.format(obj["text"], variables=variables)
        return {"type": "text", "text": text}

    @staticmethod
    def from_anthropic_block(
        obj: TextBlock,
    ) -> v1.TextContentPart:
        return v1.TextContentPart(type="text", text=obj.text)

    @staticmethod
    def from_anthropic(
        obj: TextBlockParam,
    ) -> v1.TextContentPart:
        return v1.TextContentPart(type="text", text=obj["text"])


class _ToolCallContentPartConversion:
    @staticmethod
    def to_anthropic(
        obj: v1.ToolCallContentPart,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> ToolUseBlockParam:
        id_ = obj["tool_call_id"]
        tool_call = obj["tool_call"]
        name = tool_call["name"]
        input_ = tool_call["arguments"] if "arguments" in tool_call else "{}"
        return {
            "type": "tool_use",
            "id": id_,
            "name": name,
            "input": input_,
        }

    @staticmethod
    def from_anthropic_block(
        obj: ToolUseBlock,
    ) -> v1.ToolCallContentPart:
        if isinstance(obj.input, (dict, list)):
            arguments = json.dumps(obj.input)  # pyright: ignore[reportUnknownMemberType]
        else:
            arguments = str(obj.input)
        assert isinstance(arguments, str)
        return v1.ToolCallContentPart(
            type="tool_call",
            tool_call_id=obj.id,
            tool_call=v1.ToolCallFunction(
                type="function",
                name=obj.name,
                arguments=arguments,
            ),
        )

    @staticmethod
    def from_anthropic(
        obj: ToolUseBlockParam,
    ) -> v1.ToolCallContentPart:
        if isinstance(obj["input"], (dict, list)):
            arguments = json.dumps(obj["input"])
        else:
            arguments = str(obj["input"])
        assert isinstance(arguments, str)
        return v1.ToolCallContentPart(
            type="tool_call",
            tool_call_id=obj["id"],
            tool_call=v1.ToolCallFunction(
                type="function",
                name=obj["name"],
                arguments=arguments,
            ),
        )


class _ToolResultContentPartConversion:
    @staticmethod
    def to_anthropic(
        obj: v1.ToolResultContentPart,
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
    ) -> ToolResultBlockParam:
        id_ = obj["tool_call_id"]
        param: ToolResultBlockParam = {
            "type": "tool_result",
            "tool_use_id": id_,
        }
        if isinstance(obj["tool_result"], str):
            param["content"] = obj["tool_result"]
        elif isinstance(obj["tool_result"], Sequence):
            content: list[TextBlockParam] = []
            for result in obj["tool_result"]:
                if isinstance(result, dict):
                    if (
                        "type" in result
                        and result["type"] == "text"
                        and "text" in result
                        and isinstance(result["text"], str)
                        and set(result.keys()) == {"type", "text"}  # pyright: ignore[reportUnknownArgumentType]
                    ):
                        text = str(result["text"])
                    else:
                        text = json.dumps(result)
                else:
                    text = str(result)
                content.append({"type": "text", "text": text})
            param["content"] = content
        elif isinstance(obj["tool_result"], Mapping):
            param["content"] = json.dumps(obj["tool_result"])
        elif obj["tool_result"] is not None:
            param["content"] = str(obj["tool_result"])
        return param

    @staticmethod
    def from_anthropic(
        obj: ToolResultBlockParam,
    ) -> v1.ToolResultContentPart:
        ans = v1.ToolResultContentPart(
            type="tool_result",
            tool_call_id=obj["tool_use_id"],
            tool_result=None,
        )
        if "content" in obj:
            if isinstance(obj["content"], str):
                ans["tool_result"] = obj["content"]
            elif isinstance(obj["content"], Iterable):
                ans["tool_result"] = list(obj["content"])
            elif TYPE_CHECKING:
                assert_never(obj["content"])
        return ans


class _ContentConversion:
    @overload
    @staticmethod
    def to_anthropic(
        parts: Sequence[_ContentPart],
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
        *,
        text_only: Literal[True] = True,
    ) -> Iterator[TextBlockParam]: ...

    @overload
    @staticmethod
    def to_anthropic(
        parts: Sequence[_ContentPart],
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
        *,
        text_only: Literal[False] = False,
    ) -> Iterator[_BlockParam]: ...

    @staticmethod
    def to_anthropic(
        parts: Sequence[_ContentPart],
        variables: Mapping[str, str],
        formatter: TemplateFormatter,
        /,
        *,
        text_only: bool = False,
    ) -> Any:
        for part in parts:
            if part["type"] == "text":
                yield _TextContentPartConversion.to_anthropic(part, variables, formatter)
            elif text_only:
                continue
            elif part["type"] == "tool_result":
                yield _ToolResultContentPartConversion.to_anthropic(part, variables, formatter)
            elif part["type"] == "tool_call":
                yield _ToolCallContentPartConversion.to_anthropic(part, variables, formatter)
            else:
                assert_never(part)

    @staticmethod
    def from_anthropic(
        obj: Optional[Union[str, Iterable[Union[_BlockParam, ContentBlock]]]],
    ) -> list[_ContentPart]:
        if isinstance(obj, str):
            return [
                v1.TextContentPart(
                    type="text",
                    text=obj,
                )
            ]
        content: list[_ContentPart] = []
        for block in obj or ():
            if isinstance(block, dict):
                if block["type"] == "text":
                    content.append(_TextContentPartConversion.from_anthropic(block))
                elif block["type"] == "image":
                    raise NotImplementedError
                elif block["type"] == "tool_use":
                    content.append(_ToolCallContentPartConversion.from_anthropic(block))
                elif block["type"] == "tool_result":
                    content.append(_ToolResultContentPartConversion.from_anthropic(block))
                elif block["type"] == "document":
                    raise NotImplementedError
                elif block["type"] == "thinking":
                    raise NotImplementedError
                elif block["type"] == "redacted_thinking":
                    raise NotImplementedError
                else:
                    assert_never(block["type"])
            else:
                from anthropic.types import TextBlock, ToolUseBlock

                if isinstance(block, TextBlock):
                    content.append(_TextContentPartConversion.from_anthropic_block(block))
                    continue
                if isinstance(block, ToolUseBlock):
                    content.append(_ToolCallContentPartConversion.from_anthropic_block(block))
                    continue
                if _anthropic_version < (0, 47):
                    continue
                from anthropic.types import ThinkingBlock

                if isinstance(block, ThinkingBlock):
                    raise NotImplementedError
                from anthropic.types import RedactedThinkingBlock

                if isinstance(block, RedactedThinkingBlock):
                    raise NotImplementedError
        return content


class _RoleConversion:
    @staticmethod
    def to_anthropic(
        obj: v1.PromptMessage,
    ) -> Literal["user", "assistant"]:
        role = obj["role"]
        if role == "user":
            return "user"
        if role == "assistant":
            return "assistant"
        if role == "model":
            return "assistant"
        if role == "ai":
            return "assistant"
        if role == "system":
            raise NotImplementedError
        if role == "developer":
            raise NotImplementedError
        if role == "tool":
            return "user"
        if TYPE_CHECKING:
            assert_never(role)
        return role

    @staticmethod
    def from_anthropic(
        obj: MessageParam,
    ) -> Literal["user", "assistant", "tool"]:
        if obj["role"] == "assistant":
            return "assistant"
        if obj["role"] == "user":
            if isinstance(obj["content"], list):
                for block in obj["content"]:
                    if isinstance(block, dict) and block["type"] == "tool_result":
                        return "tool"
                    else:
                        continue
            return "user"
        if TYPE_CHECKING:
            assert_never(obj["role"])
        return obj["role"]


def _get_tool_params(
    tools: Iterable[ToolUnionParam],
) -> Iterator[ToolParam]:
    for tool in tools:
        if "type" in tool:
            continue
        yield tool



from typing import Optional, Sequence, TypedDict

__all__ = ["PhoenixException", "InvalidSpanInfo", "DuplicateSpanInfo", "SpanCreationError"]


class PhoenixException(Exception):
    pass


class InvalidSpanInfo(TypedDict):
    """Information about an invalid span."""

    span_id: str
    trace_id: str
    error: str


class DuplicateSpanInfo(TypedDict):
    """Information about a duplicate span."""

    span_id: str
    trace_id: str


class SpanCreationError(PhoenixException):
    """Raised when some spans fail to be queued for creation."""

    def __init__(
        self,
        message: str,
        invalid_spans: Optional[Sequence[InvalidSpanInfo]] = None,
        duplicate_spans: Optional[Sequence[DuplicateSpanInfo]] = None,
        total_received: int = 0,
        total_queued: int = 0,
        total_invalid: int = 0,
        total_duplicates: int = 0,
    ):
        super().__init__(message)
        self.invalid_spans = invalid_spans or []
        self.duplicate_spans = duplicate_spans or []
        self.total_received = total_received
        self.total_queued = total_queued
        self.total_invalid = total_invalid
        self.total_duplicates = total_duplicates



from __future__ import annotations

import logging
from typing import Optional, cast

import httpx

from phoenix.client.__generated__ import v1
from phoenix.client.utils.encode_path_param import encode_path_param

logger = logging.getLogger(__name__)


class Projects:
    """Client for interacting with the Projects API endpoints.

    This class provides synchronous methods for creating, retrieving, updating, and deleting projects.

    Example:
        ```python
        from phoenix.client import Client

        client = Client()
        projects = client.projects.list()
        project = client.projects.get(project_id="UHJvamVjdDoy")
        ```
    """  # noqa: E501

    def __init__(self, client: httpx.Client) -> None:
        """Initialize the Projects client.

        Args:
            client: The httpx client to use for making requests.
        """
        self._client = client

    def get(
        self,
        *,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> v1.Project:
        """Get a project by ID or name.

        Args:
            project_id: The ID of the project to retrieve.
            project_name: The name of the project to retrieve.

        Returns:
            The project with the specified ID or name.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid or if neither project_id nor project_name is provided.

        Example:
            ```python
            from phoenix.client import Client

            client = Client()
            # Get by ID
            project = client.projects.get(project_id="UHJvamVjdDoy")
            # Get by name
            project = client.projects.get(project_name="My Project")
            print(f"Project name: {project['name']}")
            ```
        """  # noqa: E501
        if not project_id and not project_name:
            raise ValueError("Either project_id or project_name must be provided.")
        if project_id and project_name:
            raise ValueError("Only one of project_id or project_name can be provided.")
        if project_name:
            project_identifier = project_name
        else:
            assert project_id
            project_identifier = project_id
        url = f"v1/projects/{encode_path_param(project_identifier)}"
        response = self._client.get(url)
        response.raise_for_status()
        return cast(v1.GetProjectResponseBody, response.json())["data"]

    def list(
        self,
    ) -> list[v1.Project]:
        """List all projects.

        Returns:
            A list of all projects.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid.

        Example:
            ```python
            from phoenix.client import Client

            client = Client()
            projects = client.projects.list()
            for project in projects:
                print(f"Project name: {project['name']}")
            ```
        """  # noqa: E501
        all_projects: list[v1.Project] = []
        next_cursor: Optional[str] = None
        while True:
            url = "v1/projects"
            params = {"cursor": next_cursor} if next_cursor else {}
            response = self._client.get(url, params=params)
            response.raise_for_status()
            data = cast(v1.GetProjectsResponseBody, response.json())
            all_projects.extend(data["data"])
            if not (next_cursor := data.get("next_cursor")):
                break
        return all_projects

    def create(
        self,
        *,
        name: str,
        description: Optional[str] = None,
    ) -> v1.Project:
        """Create a new project.

        Args:
            name: The name of the project.
            description: An optional description of the project.

        Returns:
            The newly created project.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid.

        Example:
            ```python
            from phoenix.client import Client

            client = Client()
            project = client.projects.create(
                name="My Project",
                description="A description of my project",
            )
            print(f"Created project with ID: {project['id']}")
            ```
        """  # noqa: E501
        url = "v1/projects"
        json_ = v1.CreateProjectRequestBody(name=name)
        if description:
            json_["description"] = description
        response = self._client.post(url=url, json=json_)
        response.raise_for_status()
        return cast(v1.CreateProjectResponseBody, response.json())["data"]

    def update(
        self,
        *,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> v1.Project:
        """Update a project by ID or name.

        Note:
            Project names cannot be changed. If a name is provided, it will be ignored.

        Args:
            project_id: The ID of the project to update.
            project_name: The name of the project to update.
            description: The new description for the project.

        Returns:
            The updated project.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid or if neither project_id nor project_name is provided.

        Example:
            ```python
            from phoenix.client import Client

            client = Client()
            # Update by ID
            project = client.projects.update(
                project_id="UHJvamVjdDoy",
                description="Updated project description",
            )
            # Update by name
            project = client.projects.update(
                project_name="My Project",
                description="Updated project description",
            )
            print(f"Updated project description: {project['description']}")
            ```
        """  # noqa: E501
        if not project_id and not project_name:
            raise ValueError("Either project_id or project_name must be provided.")
        if project_id and project_name:
            raise ValueError("Only one of project_id or project_name can be provided.")
        if project_name:
            project_identifier = project_name
        else:
            assert project_id
            project_identifier = project_id
        url = f"v1/projects/{encode_path_param(project_identifier)}"
        if description is None:
            raise ValueError("description must be provided.")
        json_ = v1.UpdateProjectRequestBody(description=description)
        response = self._client.put(url=url, json=json_)
        response.raise_for_status()
        return cast(v1.UpdateProjectResponseBody, response.json())["data"]

    def delete(
        self,
        *,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> None:
        """Delete a project by ID or name.

        Args:
            project_id: The ID of the project to delete.
            project_name: The name of the project to delete.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If neither project_id nor project_name is provided.

        Example:
            ```python
            from phoenix.client import Client

            client = Client()
            # Delete by ID
            client.projects.delete(project_id="UHJvamVjdDoy")
            # Delete by name
            client.projects.delete(project_name="My Project")
            ```
        """  # noqa: E501
        if not project_id and not project_name:
            raise ValueError("Either project_id or project_name must be provided.")
        if project_id and project_name:
            raise ValueError("Only one of project_id or project_name can be provided.")
        if project_name:
            project_identifier = project_name
        else:
            assert project_id
            project_identifier = project_id
        url = f"v1/projects/{encode_path_param(project_identifier)}"
        response = self._client.delete(url)
        response.raise_for_status()


class AsyncProjects:
    """Asynchronous client for interacting with the Projects API endpoints.

    This class provides asynchronous methods for creating, retrieving, updating, and deleting projects.

    Example:
        ```python
        from phoenix.client import AsyncClient

        async_client = AsyncClient()
        projects = await async_client.projects.list()
        project = await async_client.projects.get(project_id="UHJvamVjdDoy")
        ```
    """  # noqa: E501

    def __init__(self, client: httpx.AsyncClient) -> None:
        """Initialize the AsyncProjects client.

        Args:
            client: The httpx async client to use for making requests.
        """
        self._client = client

    async def get(
        self,
        *,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> v1.Project:
        """Get a project by ID or name.

        Args:
            project_id: The ID of the project to retrieve.
            project_name: The name of the project to retrieve.

        Returns:
            The project with the specified ID or name.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid or if neither project_id nor project_name is provided.

        Example:
            ```python
            from phoenix.client import AsyncClient

            async_client = AsyncClient()
            project = await async_client.projects.get(project_id="UHJvamVjdDoy")
            print(f"Project name: {project['name']}")
            ```
        """  # noqa: E501
        if not project_id and not project_name:
            raise ValueError("Either project_id or project_name must be provided.")
        if project_id and project_name:
            raise ValueError("Only one of project_id or project_name can be provided.")
        if project_name:
            project_identifier = project_name
        else:
            assert project_id
            project_identifier = project_id
        url = f"v1/projects/{encode_path_param(project_identifier)}"
        response = await self._client.get(url)
        response.raise_for_status()
        return cast(v1.GetProjectResponseBody, response.json())["data"]

    async def list(
        self,
    ) -> list[v1.Project]:
        """List all projects.

        Returns:
            A list of all projects.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid.

        Example:
            ```python
            from phoenix.client import AsyncClient

            async_client = AsyncClient()
            projects = await async_client.projects.list()
            for project in projects:
                print(f"Project name: {project['name']}")
            ```
        """  # noqa: E501
        all_projects: list[v1.Project] = []
        next_cursor: Optional[str] = None
        while True:
            url = "v1/projects"
            params = {"cursor": next_cursor} if next_cursor else {}
            response = await self._client.get(url, params=params)
            response.raise_for_status()
            data = cast(v1.GetProjectsResponseBody, response.json())
            all_projects.extend(data["data"])
            if not (next_cursor := data.get("next_cursor")):
                break
        return all_projects

    async def create(
        self,
        *,
        name: str,
        description: Optional[str] = None,
    ) -> v1.Project:
        """Create a new project.

        Args:
            name: The name of the project.
            description: An optional description of the project.

        Returns:
            The newly created project.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid.

        Example:
            ```python
            from phoenix.client import AsyncClient

            async_client = AsyncClient()
            project = await async_client.projects.create(
                name="My Project",
                description="A description of my project",
            )
            print(f"Created project with ID: {project['id']}")
            ```
        """  # noqa: E501
        url = "v1/projects"
        json_ = v1.CreateProjectRequestBody(name=name)
        if description:
            json_["description"] = description
        response = await self._client.post(url=url, json=json_)
        response.raise_for_status()
        return cast(v1.CreateProjectResponseBody, response.json())["data"]

    async def update(
        self,
        *,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> v1.Project:
        """Update a project by ID or name.

        Note:
            Project names cannot be changed. If a name is provided, it will be ignored.

        Args:
            project_id: The ID of the project to update.
            project_name: The name of the project to update.
            description: The new description for the project.

        Returns:
            The updated project.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid or if neither project_id nor project_name is provided.

        Example:
            ```python
            from phoenix.client import AsyncClient

            async_client = AsyncClient()
            # Update by ID
            project = await async_client.projects.update(
                project_id="UHJvamVjdDoy",
                description="Updated project description",
            )
            # Update by name
            project = await async_client.projects.update(
                project_name="My Project",
                description="Updated project description",
            )
            print(f"Updated project description: {project['description']}")
            ```
        """  # noqa: E501
        if not project_id and not project_name:
            raise ValueError("Either project_id or project_name must be provided.")
        if project_id and project_name:
            raise ValueError("Only one of project_id or project_name can be provided.")
        if project_name:
            project_identifier = project_name
        else:
            assert project_id
            project_identifier = project_id
        url = f"v1/projects/{encode_path_param(project_identifier)}"
        if description is None:
            raise ValueError("description must be provided.")
        json_ = v1.UpdateProjectRequestBody(description=description)
        response = await self._client.put(url=url, json=json_)
        response.raise_for_status()
        return cast(v1.UpdateProjectResponseBody, response.json())["data"]

    async def delete(
        self,
        *,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> None:
        """Delete a project by ID or name.

        Args:
            project_id: The ID of the project to delete.
            project_name: The name of the project to delete.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If neither project_id nor project_name is provided.

        Example:
            ```python
            from phoenix.client import AsyncClient

            async_client = AsyncClient()
            # Delete by ID
            await async_client.projects.delete(project_id="UHJvamVjdDoy")
            # Delete by name
            await async_client.projects.delete(project_name="My Project")
            ```
        """  # noqa: E501
        if not project_id and not project_name:
            raise ValueError("Either project_id or project_name must be provided.")
        if project_id and project_name:
            raise ValueError("Only one of project_id or project_name can be provided.")
        if project_name:
            project_identifier = project_name
        else:
            assert project_id
            project_identifier = project_id
        url = f"v1/projects/{encode_path_param(project_identifier)}"
        response = await self._client.delete(url)
        response.raise_for_status()



from __future__ import annotations

from typing import Mapping, Optional

import httpx

from phoenix.client.resources.annotations import Annotations, AsyncAnnotations
from phoenix.client.resources.datasets import AsyncDatasets, Datasets
from phoenix.client.resources.experiments import AsyncExperiments, Experiments
from phoenix.client.resources.projects import AsyncProjects, Projects
from phoenix.client.resources.prompts import AsyncPrompts, Prompts
from phoenix.client.resources.spans import AsyncSpans, Spans
from phoenix.client.utils.config import get_base_url, get_env_client_headers


class Client:
    def __init__(
        self,
        *,
        base_url: str | httpx.URL | None = None,
        api_key: str | None = None,
        headers: Mapping[str, str] | None = None,
        http_client: httpx.Client | None = None,
    ):
        """
        Initializes a Client instance.

        Args:
            base_url (Optional[str]): The base URL for the API endpoint. If not provided, it will
                be read from the environment variables or fall back to http://localhost:6006/.
            api_key (Optional[str]): The API key for authentication. If provided, it will be
                included in the Authorization header as a bearer token. Defaults to None.
            headers (Optional[Mapping[str, str]]): Additional headers to be included in the HTTP.
                Defaults to None. This is ignored if http_client is provided. Additional headers
                may be added from the environment variables, but won't override specified values.
            http_client (Optional[httpx.Client]): An instance of httpx.Client to be used for
                making HTTP requests. If not provided, a new instance will be created. Defaults
                to None.
        """  # noqa: E501
        if http_client is None:
            base_url = base_url or get_base_url()
            self._client = _WrappedClient(
                base_url=base_url,
                headers=_update_headers(headers, api_key),
            )
        else:
            self._client = http_client

    @property
    def _client(self) -> httpx.Client:
        return self._http_client

    @_client.setter
    def _client(self, value: httpx.Client) -> None:
        self._http_client = value
        self._prompts = Prompts(value)
        self._projects = Projects(value)
        self._spans = Spans(value)
        self._annotations = Annotations(value)
        self._datasets = Datasets(value)
        self._experiments = Experiments(value)

    @property
    def prompts(self) -> Prompts:
        """
        Returns an instance of the Prompts class for interacting with prompt-related API endpoints.

        Returns:
            Prompts: An instance of the Prompts class.
        """  # noqa: E501
        return self._prompts

    @property
    def projects(self) -> Projects:
        """
        Returns an instance of the Projects class for interacting with project-related API endpoints.

        Returns:
            Projects: An instance of the Projects class.
        """  # noqa: E501
        return self._projects

    @property
    def spans(self) -> Spans:
        """
        Returns an instance of the Spans class for interacting with span-related
        API endpoints.

        Returns:
            Spans: An instance of the Spans class.
        """
        return self._spans

    @property
    def annotations(self) -> Annotations:
        """
        Returns an instance of the Annotations class for interacting with annotation-related
        API endpoints.

        Returns:
            Annotations: An instance of the Annotations class.
        """  # noqa: E501
        return self._annotations

    @property
    def datasets(self) -> Datasets:
        """
        Returns an instance of the Datasets class for interacting with dataset-related
        API endpoints.
        """
        return self._datasets

    @property
    def experiments(self) -> Experiments:
        """
        Returns an instance of the Experiments class for interacting with experiment-related
        API endpoints.
        """
        return self._experiments


class AsyncClient:
    def __init__(
        self,
        *,
        base_url: str | httpx.URL | None = None,
        api_key: str | None = None,
        headers: Mapping[str, str] | None = None,
        http_client: httpx.AsyncClient | None = None,
    ):
        """
        Initializes an Asynchronous Client instance.

        Args:
            base_url (Optional[str]): The base URL for the API endpoint. If not provided, it will
                be read from the environment variables or fall back to http://localhost:6006/.
            api_key (Optional[str]): The API key for authentication. If provided, it will be
                included in the Authorization header as a bearer token. Defaults to None.
            headers (Optional[Mapping[str, str]]): Additional headers to be included in the HTTP.
                Defaults to None. This is ignored if http_client is provided. Additional headers
                may be added from the environment variables, but won't override specified values.
            http_client (Optional[httpx.AsyncClient]): An instance of httpx.AsyncClient to be used
                for making HTTP requests. If not provided, a new instance will be created. Defaults
                to None.
        """  # noqa: E501
        if http_client is None:
            base_url = base_url or get_base_url()
            http_client = httpx.AsyncClient(
                base_url=base_url,
                headers=_update_headers(headers, api_key),
            )
        self._client = http_client

    @property
    def _client(self) -> httpx.AsyncClient:
        return self._http_client

    @_client.setter
    def _client(self, value: httpx.AsyncClient) -> None:
        self._http_client = value
        self._prompts = AsyncPrompts(value)
        self._projects = AsyncProjects(value)
        self._spans = AsyncSpans(value)
        self._annotations = AsyncAnnotations(value)
        self._datasets = AsyncDatasets(value)
        self._experiments = AsyncExperiments(value)

    @property
    def prompts(self) -> AsyncPrompts:
        """
        Returns an instance of the Asynchronous Prompts class for interacting with prompt-related
        API endpoints.

        Returns:
            AsyncPrompts: An instance of the Prompts class.
        """  # noqa: E501
        return self._prompts

    @property
    def projects(self) -> AsyncProjects:
        """
        Returns an instance of the Asynchronous Projects class for interacting with project-related
        API endpoints.

        Returns:
            AsyncProjects: An instance of the Projects class.
        """  # noqa: E501
        return self._projects

    @property
    def spans(self) -> AsyncSpans:
        """
        Returns an instance of the Asynchronous Spans class for interacting with span-related
        API endpoints.

        Returns:
            AsyncSpans: An instance of the Spans class.
        """
        return self._spans

    @property
    def annotations(self) -> AsyncAnnotations:
        """
        Returns an instance of the Asynchronous Annotations class for interacting with annotation-related
        API endpoints.

        Returns:
            AsyncAnnotations: An instance of the Annotations class.
        """  # noqa: E501
        return self._annotations

    @property
    def datasets(self) -> AsyncDatasets:
        """
        Returns an instance of the Asynchronous Datasets class for interacting with dataset-related
        API endpoints.
        """
        return self._datasets

    @property
    def experiments(self) -> AsyncExperiments:
        """
        Returns an instance of the Asynchronous Experiments class for interacting with
        experiment-related API endpoints.
        """
        return self._experiments


def _update_headers(
    headers: Optional[Mapping[str, str]],
    api_key: Optional[str],
) -> dict[str, str]:
    headers = dict(headers or {})
    for k, v in get_env_client_headers().items():
        if k not in headers:
            headers[k] = v
    if api_key:
        headers = {
            **{k: v for k, v in (headers or {}).items() if k.lower() != "authorization"},
            "Authorization": f"Bearer {api_key}",
        }
    return headers


class _WrappedClient(httpx.Client):
    def __del__(self) -> None:
        try:
            self.close()
        except BaseException:
            pass



import os
from typing import Optional, overload

import httpx

from phoenix.client.constants import (
    ENV_OTEL_EXPORTER_OTLP_ENDPOINT,
    ENV_PHOENIX_API_KEY,
    ENV_PHOENIX_CLIENT_HEADERS,
    ENV_PHOENIX_COLLECTOR_ENDPOINT,
    ENV_PHOENIX_HOST,
    ENV_PHOENIX_HOST_ROOT_PATH,
    ENV_PHOENIX_PORT,
    HOST,
    PORT,
)
from phoenix.client.utils.parse_env_headers import parse_env_headers


def get_env_phoenix_api_key() -> Optional[str]:
    return getenv(ENV_PHOENIX_API_KEY)


def get_env_port() -> int:
    if not (port := getenv(ENV_PHOENIX_PORT)):
        return PORT
    if port.isnumeric():
        return int(port)
    raise ValueError(
        f"Invalid value for environment variable {ENV_PHOENIX_PORT}: "
        f"{port}. Value must be an integer."
    )


def get_env_host() -> str:
    return getenv(ENV_PHOENIX_HOST) or HOST


def get_env_host_root_path() -> str:
    if (host_root_path := getenv(ENV_PHOENIX_HOST_ROOT_PATH)) is None:
        return ""
    if not host_root_path.startswith("/"):
        raise ValueError(
            f"Invalid value for environment variable {ENV_PHOENIX_HOST_ROOT_PATH}: "
            f"{host_root_path}. Value must start with '/'"
        )
    if host_root_path.endswith("/"):
        raise ValueError(
            f"Invalid value for environment variable {ENV_PHOENIX_HOST_ROOT_PATH}: "
            f"{host_root_path}. Value cannot end with '/'"
        )
    return host_root_path


def get_env_client_headers() -> dict[str, str]:
    headers = parse_env_headers(getenv(ENV_PHOENIX_CLIENT_HEADERS))
    if (api_key := get_env_phoenix_api_key()) and "authorization" not in [
        k.lower() for k in headers
    ]:
        headers["Authorization"] = f"Bearer {api_key}"
    return headers


def get_env_collector_endpoint() -> Optional[str]:
    return getenv(ENV_PHOENIX_COLLECTOR_ENDPOINT) or getenv(ENV_OTEL_EXPORTER_OTLP_ENDPOINT)


def get_base_url() -> httpx.URL:
    host: str = get_env_host()
    if host == "0.0.0.0":
        host = "127.0.0.1"
    base_url: str = get_env_collector_endpoint() or f"http://{host}:{get_env_port()}"
    return httpx.URL(base_url)


@overload
def getenv(key: str) -> Optional[str]: ...
@overload
def getenv(key: str, default: str) -> str: ...
def getenv(key: str, default: Optional[str] = None) -> Optional[str]:
    """
    Retrieves the value of an environment variable.

    Parameters
    ----------
        key : str
            The name of the environment variable.
        default : Optional[str], optional
            The default value to return if the environment variable is not set, by default None.

    Returns
    -------
    Optional[str]
        The value of the environment variable, or `default` if the variable is not set.
        Leading and trailing whitespaces are stripped from the value, assuming they were
        inadvertently added.
    """
    if (value := os.getenv(key)) is None:
        return default
    return value.strip()



# pyright: reportUnknownMemberType=false
import json
from secrets import token_hex
from typing import Any, Mapping

import pytest
from deepdiff.diff import DeepDiff
from faker import Faker
from google.generativeai import protos
from google.generativeai.types import content_types

from phoenix.client.__generated__.v1 import TextContentPart
from phoenix.client.helpers.sdk.google_generativeai.generate_content import (
    _ContentConversion,
    _FunctionDeclarationConversion,
    _TextContentPartConversion,
    _ToolKwargs,
    _ToolKwargsConversion,
)
from phoenix.client.utils.template_formatters import NO_OP_FORMATTER


def _dict() -> dict[str, Any]:
    return Faker().pydict(3, value_types=(int, float, bool, str))  # pyright: ignore[reportUnknownMemberType]


def _text() -> protos.Part:
    ans = protos.Part()  # type: ignore[no-untyped-call]
    ans.text = token_hex(8)
    return ans


class TestContentConversion:
    @pytest.mark.parametrize(
        "obj",
        [
            protos.Content(role="user", parts=[_text(), _text()]),  # type: ignore[no-untyped-call]
        ],
    )
    def test_round_trip(self, obj: protos.Content) -> None:
        new_obj: protos.Content = next(
            _ContentConversion.to_google(_ContentConversion.from_google(obj), {}, NO_OP_FORMATTER)
        )
        assert not DeepDiff(
            protos.Content.to_dict(obj),
            protos.Content.to_dict(new_obj),
        )


class TestTextPartConversion:
    def test_round_trip(self) -> None:
        obj: protos.Part = _text()
        new_obj: protos.Part = _TextContentPartConversion.to_google(
            _TextContentPartConversion.from_google(obj), {}, NO_OP_FORMATTER
        )
        assert not DeepDiff(
            protos.Part.to_dict(obj),
            protos.Part.to_dict(new_obj),
        )

    def test_formatter(self) -> None:
        obj = TextContentPart(type="text", text=token_hex(8))
        formatter, variables = _MockFormatter(), _dict()
        part: protos.Part = _TextContentPartConversion.to_google(obj, variables, formatter)
        assert part.text == formatter.format(obj["text"], variables=variables)


def _f(a: int, b: list[float], c: str, d: list[bool], e: dict[str, Any]) -> int:
    raise NotImplementedError


def _g(a: list[int], b: float, c: list[str], d: bool, e: list[dict[str, Any]]) -> str:
    raise NotImplementedError


_FUNCTION_DECLARATIONS = [
    content_types.FunctionDeclaration.from_function(_f),
    content_types.FunctionDeclaration.from_function(_g),
]


class TestFunctionDeclarationConversion:
    @pytest.mark.parametrize(
        "obj",
        _FUNCTION_DECLARATIONS,
    )
    def test_round_trip(self, obj: content_types.FunctionDeclaration) -> None:
        new_obj: content_types.FunctionDeclaration = _FunctionDeclarationConversion.to_google(
            _FunctionDeclarationConversion.from_google(obj)
        )
        assert not DeepDiff(
            protos.FunctionDeclaration.to_dict(obj.to_proto()),
            protos.FunctionDeclaration.to_dict(new_obj.to_proto()),
        )


_TOOLS = [content_types.Tool(function_declarations=_FUNCTION_DECLARATIONS)]


class TestToolKwargsConversion:
    @pytest.mark.parametrize(
        "obj",
        [
            {
                "tools": _TOOLS,
                "tool_config": protos.ToolConfig(  # type: ignore[no-untyped-call]
                    function_calling_config=protos.FunctionCallingConfig(  # type: ignore[no-untyped-call]
                        mode=protos.FunctionCallingConfig.Mode.ANY
                    ),
                ),
            },
            {
                "tools": _TOOLS,
                "tool_config": protos.ToolConfig(  # type: ignore[no-untyped-call]
                    function_calling_config=protos.FunctionCallingConfig(  # type: ignore[no-untyped-call]
                        mode=protos.FunctionCallingConfig.Mode.NONE
                    ),
                ),
            },
            {
                "tools": _TOOLS,
                "tool_config": protos.ToolConfig(  # type: ignore[no-untyped-call]
                    function_calling_config=protos.FunctionCallingConfig(  # type: ignore[no-untyped-call]
                        mode=protos.FunctionCallingConfig.Mode.AUTO
                    ),
                ),
            },
        ],
    )
    def test_round_trip(self, obj: _ToolKwargs) -> None:
        new_obj = _ToolKwargsConversion.to_google(_ToolKwargsConversion.from_google(obj))
        assert "tools" in obj
        assert "tools" in new_obj
        for i in range(len(obj["tools"])):
            assert not DeepDiff(
                protos.Tool.to_dict(obj["tools"][i].to_proto()),  # type: ignore[no-untyped-call]
                protos.Tool.to_dict(new_obj["tools"][i].to_proto()),  # type: ignore[no-untyped-call]
            )
        assert "tool_config" in obj
        assert "tool_config" in new_obj
        assert not DeepDiff(
            protos.ToolConfig.to_dict(obj["tool_config"]),
            protos.ToolConfig.to_dict(new_obj["tool_config"]),
        )


class _MockFormatter:
    def format(self, _: str, /, *, variables: Mapping[str, str]) -> str:
        return json.dumps(variables)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Iterable, Iterator, Literal, Optional, cast, get_args

import httpx
from typing_extensions import TypeAlias

from phoenix.client.__generated__ import v1

if TYPE_CHECKING:
    import pandas as pd

logger = logging.getLogger(__name__)

_AnnotatorKind: TypeAlias = Literal["LLM", "CODE", "HUMAN"]
_VALID_ANNOTATOR_KINDS: frozenset[_AnnotatorKind] = frozenset(get_args(_AnnotatorKind))
_DATAFRAME_CHUNK_SIZE = 100


class Annotations:
    """Client for interacting with the Annotations API endpoints.

    This class provides synchronous methods for creating and managing span annotations.

    Example:
        ```python
        from phoenix.client import Client

        client = Client()
        annotation = client.annotations.add_span_annotation(
            annotation_name="sentiment",
            span_id="abc123",
            label="positive",
            score=0.9,
        )
        ```
    """  # noqa: E501

    def __init__(self, client: httpx.Client) -> None:
        """Initialize the Annotations client.

        Args:
            client: The httpx client to use for making requests.
        """
        self._client = client

    def add_span_annotation(
        self,
        *,
        span_id: str,
        annotation_name: str,
        annotator_kind: Literal["LLM", "CODE", "HUMAN"] = "HUMAN",
        label: Optional[str] = None,
        score: Optional[float] = None,
        explanation: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        identifier: Optional[str] = None,
        sync: bool = False,
    ) -> Optional[v1.InsertedSpanAnnotation]:
        """Add a single span annotation.

        Args:
            annotation_name: The name of the annotation.
            span_id: The ID of the span to annotate.
            annotator_kind: The kind of annotator used for the annotation. Must be one of "LLM", "CODE", or "HUMAN".
            label: The label assigned by the annotation.
            score: The score assigned by the annotation.
            explanation: Explanation of the annotation result.
            metadata: Additional metadata for the annotation.
            identifier: An optional identifier for the annotation. Each annotation is uniquely identified by the combination
                of name, span_id, and identifier (where a null identifier is equivalent to an empty string).
                If an annotation with the same name, span_id, and identifier already exists, it will be updated.
                Using a non-empty identifier allows you to have multiple annotations with the same name and span_id.
                Most of the time, you can leave this as None - it will also update the record with identifier="" if it exists.
            sync: If True, the request will be fulfilled synchronously and the response will contain
                the inserted annotation ID. If False, the request will be processed asynchronously.

        Returns:
            If sync is True, the inserted span annotation containing an ID. If sync is False, None.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid or if at least one of label, score, or explanation
                is not provided.

        Example:
            ```python
            from phoenix.client import Client

            client = Client()
            client.annotations.add_span_annotation(
                annotation_name="sentiment",
                span_id="abc123",
                label="positive",
                score=0.9,
                explanation="The text expresses a positive sentiment.",
                sync=True,
            )
            ```
        """  # noqa: E501
        anno = _get_span_annotation(
            span_id=span_id,
            annotation_name=annotation_name,
            annotator_kind=annotator_kind,
            label=label,
            score=score,
            explanation=explanation,
            metadata=metadata,
            identifier=identifier,
        )
        if res := self.log_span_annotations(span_annotations=[anno], sync=sync):
            return res[0]
        return None

    def log_span_annotations_dataframe(
        self,
        *,
        dataframe: pd.DataFrame,
        annotator_kind: Optional[Literal["LLM", "CODE", "HUMAN"]] = None,
        annotation_name: Optional[str] = None,
        sync: bool = False,
    ) -> Optional[list[v1.InsertedSpanAnnotation]]:
        """Log multiple span annotations from a pandas DataFrame.

        This method allows you to create multiple span annotations at once by providing the data in a pandas DataFrame.
        The DataFrame can either include `name` or `annotation_name` columns (but not both) and `annotator_kind` column,
        or you can specify global values for all rows. The data is processed in chunks of 100 rows for efficient batch processing.

        Args:
            dataframe: A pandas DataFrame containing the annotation data. Must include either a "name" or "annotation_name" column
                (but not both) or provide a global annotation_name parameter. Similarly, must include an "annotator_kind" column
                or provide a global annotator_kind. The `span_id` can be either a column in the DataFrame or will be taken from
                the DataFrame index. Optional columns include: "label", "score", "explanation", "metadata", and "identifier".
            annotator_kind: Optional. The kind of annotator used for all annotations. If provided, this value will be used
                for all rows and the DataFrame does not need to include an "annotator_kind" column.
                Must be one of "LLM", "CODE", or "HUMAN".
            annotation_name: Optional. The name to use for all annotations. If provided, this value will be used
                for all rows and the DataFrame does not need to include a "name" or "annotation_name" column.
            sync: If True, the request will be fulfilled synchronously and the response will contain
                the inserted annotation IDs. If False, the request will be processed asynchronously.

        Returns:
            If sync is True, a list of all inserted span annotations. If sync is False, None.

        Raises:
            ImportError: If pandas is not installed.
            ValueError: If the DataFrame is missing required columns, if both "name" and "annotation_name" columns are present,
                or if no valid annotation data is provided.

        Example:
            ```python
            import pandas as pd

            # Using name and annotator_kind from DataFrame
            df1 = pd.DataFrame({
                "name": ["sentiment", "toxicity"],
                "annotator_kind": ["HUMAN", "LLM"],
                "label": ["positive", "low"],
                "score": [0.9, 0.1]
            })
            client.annotations.log_span_annotations_dataframe(dataframe=df1)

            # Using annotation_name and annotator_kind from DataFrame
            df2 = pd.DataFrame({
                "annotation_name": ["sentiment", "toxicity"],
                "annotator_kind": ["HUMAN", "LLM"],
                "label": ["positive", "low"],
                "score": [0.9, 0.1]
            })
            client.annotations.log_span_annotations_dataframe(dataframe=df2)

            # Using global name and annotator_kind
            df3 = pd.DataFrame({
                "label": ["positive", "low"]
            }, index=["span1", "span2"])
            client.annotations.log_span_annotations_dataframe(
                dataframe=df3,
                annotation_name="sentiment",  # applies to all rows
                annotator_kind="HUMAN"  # applies to all rows
            )
            ```
        """  # noqa: E501
        # Process DataFrame chunks using iterator
        all_responses: list[v1.InsertedSpanAnnotation] = []
        for chunk in _chunk_dataframe(
            dataframe=dataframe,
            annotation_name=annotation_name,
            annotator_kind=annotator_kind,
            chunk_size=_DATAFRAME_CHUNK_SIZE,
        ):
            # Delegate to log_span_annotations
            response = self.log_span_annotations(span_annotations=chunk, sync=sync)
            if sync and response:
                all_responses.extend(response)

        return all_responses if sync else None

    def log_span_annotations(
        self,
        *,
        span_annotations: Iterable[v1.SpanAnnotationData],
        sync: bool = False,
    ) -> Optional[list[v1.InsertedSpanAnnotation]]:
        """Log multiple span annotations.

        Args:
            span_annotations: An iterable of span annotation data to log. Each annotation must include
                at least a span_id, name, and annotator_kind, and at least one of label, score, or explanation.
            sync: If True, the request will be fulfilled synchronously and the response will contain
                the inserted annotation IDs. If False, the request will be processed asynchronously.

        Returns:
            If sync is True, a list of inserted span annotations, each containing an ID. If sync is False, None.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid or if the input is invalid.
        """  # noqa: E501
        # Convert to list and validate input
        annotations_list = list(span_annotations)
        if not annotations_list:
            raise ValueError("span_annotations cannot be empty")

        url = "v1/span_annotations"
        params = {"sync": sync} if sync else {}
        json_ = v1.AnnotateSpansRequestBody(data=annotations_list)
        response = self._client.post(url=url, json=json_, params=params)
        response.raise_for_status()
        if not sync:
            return None
        return list(cast(v1.AnnotateSpansResponseBody, response.json())["data"])


class AsyncAnnotations:
    """Asynchronous client for interacting with the Annotations API endpoints.

    This class provides asynchronous methods for creating and managing span annotations.

    Example:
        ```python
        from phoenix.client import AsyncClient

        async_client = AsyncClient()
        annotation = await async_client.annotations.add_span_annotation(
            annotation_name="sentiment",
            span_id="abc123",
            label="positive",
            score=0.9,
        )
        ```
    """  # noqa: E501

    def __init__(self, client: httpx.AsyncClient) -> None:
        """Initialize the AsyncAnnotations client.

        Args:
            client: The httpx async client to use for making requests.
        """
        self._client = client

    async def add_span_annotation(
        self,
        *,
        span_id: str,
        annotation_name: str,
        annotator_kind: Literal["LLM", "CODE", "HUMAN"] = "HUMAN",
        label: Optional[str] = None,
        score: Optional[float] = None,
        explanation: Optional[str] = None,
        metadata: Optional[dict[str, Any]] = None,
        identifier: Optional[str] = None,
        sync: bool = False,
    ) -> Optional[v1.InsertedSpanAnnotation]:
        """Add a single span annotation asynchronously.

        Args:
            annotation_name: The name of the annotation.
            span_id: The ID of the span to annotate.
            annotator_kind: The kind of annotator used for the annotation. Must be one of "LLM", "CODE", or "HUMAN".
            label: The label assigned by the annotation.
            score: The score assigned by the annotation.
            explanation: Explanation of the annotation result.
            metadata: Additional metadata for the annotation.
            identifier: An optional identifier for the annotation. Each annotation is uniquely identified by the combination
                of name, span_id, and identifier (where a null identifier is equivalent to an empty string).
                If an annotation with the same name, span_id, and identifier already exists, it will be updated.
                Using a non-empty identifier allows you to have multiple annotations with the same name and span_id.
                Most of the time, you can leave this as None - it will also update the record with identifier="" if it exists.
            sync: If True, the request will be fulfilled synchronously and the response will contain
                the inserted annotation ID. If False, the request will be processed asynchronously.

        Returns:
            If sync is True, the inserted span annotation containing an ID. If sync is False, None.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid or if at least one of label, score, or explanation
                is not provided.

        Example:
            ```python
            from phoenix.client import AsyncClient

            async_client = AsyncClient()
            await async_client.annotations.add_span_annotation(
                annotation_name="sentiment",
                span_id="abc123",
                label="positive",
                score=0.9,
                explanation="The text expresses a positive sentiment.",
                sync=True,
            )
            ```
        """  # noqa: E501
        anno = _get_span_annotation(
            span_id=span_id,
            annotation_name=annotation_name,
            annotator_kind=annotator_kind,
            label=label,
            score=score,
            explanation=explanation,
            metadata=metadata,
            identifier=identifier,
        )
        if res := await self.log_span_annotations(span_annotations=[anno], sync=sync):
            return res[0]
        return None

    async def log_span_annotations_dataframe(
        self,
        *,
        dataframe: pd.DataFrame,
        annotation_name: Optional[str] = None,
        annotator_kind: Optional[Literal["LLM", "CODE", "HUMAN"]] = None,
        sync: bool = False,
    ) -> Optional[list[v1.InsertedSpanAnnotation]]:
        """Log multiple span annotations from a pandas DataFrame asynchronously.

        This method allows you to create multiple span annotations at once by providing the data in a pandas DataFrame.
        The DataFrame can either include `name` and `annotator_kind` columns or you can specify global values for all rows.
        The data is processed in chunks of 100 rows for efficient batch processing.

        Args:
            dataframe: A pandas DataFrame containing the annotation data. Must include either a "name" column or provide
                a global name parameter. Similarly, must include an "annotator_kind" column or provide a global annotator_kind.
                The `span_id` can be either a column in the DataFrame or will be taken from the DataFrame index.
                Optional columns include: "label", "score", "explanation", "metadata", and "identifier".
            annotator_kind: Optional. The kind of annotator used for all annotations. If provided, this value will be used
                for all rows and the DataFrame does not need to include an "annotator_kind" column.
                Must be one of "LLM", "CODE", or "HUMAN".
            annotation_name: Optional. The name to use for all annotations. If provided, this value will be used
                for all rows and the DataFrame does not need to include a "name" column.
            sync: If True, the request will be fulfilled synchronously and the response will contain
                the inserted annotation IDs. If False, the request will be processed asynchronously.

        Returns:
            If sync is True, a list of all inserted span annotations. If sync is False, None.

        Raises:
            ImportError: If pandas is not installed.
            ValueError: If the DataFrame is missing required columns or if no valid annotation data is provided.

        Example:
            ```python
            import pandas as pd

            # Using name and annotator_kind from DataFrame
            df1 = pd.DataFrame({
                "name": ["sentiment", "toxicity"],
                "annotator_kind": ["HUMAN", "LLM"],
                "label": ["positive", "low"],
                "score": [0.9, 0.1]
            })
            await async_client.annotations.log_span_annotations_dataframe(dataframe=df1)

            # Using global name and annotator_kind
            df2 = pd.DataFrame({
                "label": ["positive", "low"]
            }, index=["span1", "span2"])
            await async_client.annotations.log_span_annotations_dataframe(
                dataframe=df2,
                annotation_name="sentiment",  # applies to all rows
                annotator_kind="HUMAN"  # applies to all rows
            )
            ```
        """  # noqa: E501
        # Process DataFrame chunks using iterator
        all_responses: list[v1.InsertedSpanAnnotation] = []
        for chunk in _chunk_dataframe(
            dataframe=dataframe,
            annotation_name=annotation_name,
            annotator_kind=annotator_kind,
            chunk_size=_DATAFRAME_CHUNK_SIZE,
        ):
            # Delegate to log_span_annotations
            response = await self.log_span_annotations(span_annotations=chunk, sync=sync)
            if sync and response:
                all_responses.extend(response)

        return all_responses if sync else None

    async def log_span_annotations(
        self,
        *,
        span_annotations: Iterable[v1.SpanAnnotationData],
        sync: bool = False,
    ) -> Optional[list[v1.InsertedSpanAnnotation]]:
        """Log multiple span annotations asynchronously.

        Args:
            span_annotations: An iterable of span annotation data to log. Each annotation must include
                at least a span_id, name, and annotator_kind, and at least one of label, score, or explanation.
            sync: If True, the request will be fulfilled synchronously and the response will contain
                the inserted annotation IDs. If False, the request will be processed asynchronously.

        Returns:
            If sync is True, a list of inserted span annotations, each containing an ID. If sync is False, None.

        Raises:
            httpx.HTTPError: If the request fails.
            ValueError: If the response is invalid.

        Example:
            ```python
            from phoenix.client import AsyncClient

            async_client = AsyncClient()

            # Create span annotation data objects
            annotation1 = {
                "name": "sentiment",
                "span_id": "span_123",
                "annotator_kind": "HUMAN",
                "result": {
                    "label": "positive",
                    "score": 0.9
                },
                "metadata": {"source": "user_feedback"}
            }

            annotation2 = {
                "name": "toxicity",
                "span_id": "span_456",
                "annotator_kind": "LLM",
                "result": {
                    "label": "low",
                    "score": 0.1,
                    "explanation": "No harmful content detected"
                }
            }

            # Log multiple annotations at once
            await async_client.annotations.log_span_annotations(
                span_annotations=[annotation1, annotation2],
            )
            ```
        """  # noqa: E501
        # Convert to list and validate input
        annotations_list: list[v1.SpanAnnotationData] = list(span_annotations)
        if not annotations_list:
            raise ValueError("span_annotations cannot be empty")

        url = "v1/span_annotations"
        params = {"sync": sync} if sync else {}
        json_ = v1.AnnotateSpansRequestBody(data=annotations_list)
        response = await self._client.post(url=url, json=json_, params=params)
        response.raise_for_status()
        if not sync:
            return None
        return list(cast(v1.AnnotateSpansResponseBody, response.json())["data"])


def _get_span_annotation(
    *,
    span_id: str,
    annotation_name: str,
    annotator_kind: Literal["LLM", "CODE", "HUMAN"] = "HUMAN",
    label: Optional[str] = None,
    score: Optional[float] = None,
    explanation: Optional[str] = None,
    metadata: Optional[dict[str, Any]] = None,
    identifier: Optional[str] = None,
) -> v1.SpanAnnotationData:
    """Create a span annotation data object.

    Args:
        annotation_name: The name of the annotation.
        span_id: The ID of the span to annotate.
        annotator_kind: The kind of annotator used for the annotation. Must be one of "LLM", "CODE", or "HUMAN".
        label: The label assigned by the annotation.
        score: The score assigned by the annotation.
        explanation: Explanation of the annotation result.
        metadata: Additional metadata for the annotation.
        identifier: An optional identifier for the annotation. Each annotation is uniquely identified by the combination
            of name, span_id, and identifier (where a null identifier is equivalent to an empty string).
            If an annotation with the same name, span_id, and identifier already exists, it will be updated.
            Using a non-empty identifier allows you to have multiple annotations with the same name and span_id.
            Most of the time, you can leave this as None - it will still update the record if it exists.
            It will also update the record with identifier="" if it exists.

    Returns:
        A span annotation data object that can be used with the Annotations API.

    Raises:
        ValueError: If at least one of label, score, or explanation is not provided, or if required fields are invalid.
    """  # noqa: E501
    # Validate required fields
    if not span_id or not isinstance(span_id, str):  # pyright: ignore[reportUnnecessaryIsInstance]
        raise ValueError("span_id must be a non-empty string")
    if not annotation_name or not isinstance(annotation_name, str):  # pyright: ignore[reportUnnecessaryIsInstance]
        raise ValueError("annotation_name must be a non-empty string")
    if annotator_kind not in _VALID_ANNOTATOR_KINDS:
        raise ValueError(f"annotator_kind must be one of {_VALID_ANNOTATOR_KINDS}")

    # Validate that at least one of label, score, or explanation is provided
    if not label and score is None and not explanation:
        raise ValueError("At least one of label, score, or explanation must be provided.")

    # Validate score if provided
    if score is not None and not isinstance(score, (int, float)):  # pyright: ignore[reportUnnecessaryIsInstance]
        raise ValueError("score must be a number")

    # Validate metadata if provided
    if metadata is not None and not isinstance(metadata, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
        raise ValueError("metadata must be a dictionary")

    result: v1.SpanAnnotationResult = {}
    if label:
        result["label"] = label
    if score is not None:
        result["score"] = score
    if explanation:
        result["explanation"] = explanation
    anno = v1.SpanAnnotationData(
        name=annotation_name,
        span_id=span_id,
        annotator_kind=annotator_kind,
        result=result,
    )
    if metadata:
        anno["metadata"] = metadata
    if identifier and identifier.strip():
        anno["identifier"] = identifier.strip()
    return anno


def _validate_dataframe(
    *,
    dataframe: pd.DataFrame,
    annotation_name: Optional[str] = None,
    annotator_kind: Optional[Literal["LLM", "CODE", "HUMAN"]] = None,
) -> None:
    """Internal function to validate that the DataFrame has the required columns and data.

    This function performs comprehensive validation of the DataFrame structure and content,
    including type checking, required columns, and value validation.

    Args:
        dataframe: The DataFrame to validate
        annotation_name: Optional global name value. If provided, must be a non-empty string.
        annotator_kind: Optional global annotator_kind value. Must be one of "LLM", "CODE", or "HUMAN".

    Raises:
        ValueError: If the DataFrame is missing required columns, if no valid annotation data is provided,
            or if annotator_kind values are invalid.
        TypeError: If the input is not a pandas DataFrame.
    """  # noqa: E501
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "Pandas is not installed. Please install pandas to use this method: "
            "pip install pandas"
        )

    # Type check for DataFrame
    if not isinstance(dataframe, pd.DataFrame):  # pyright: ignore[reportUnnecessaryIsInstance]
        raise TypeError(f"Expected pandas DataFrame, got {type(dataframe)}")

    # Check if DataFrame is empty
    if dataframe.empty:
        raise ValueError("DataFrame cannot be empty")

    # Validate global name if provided
    if annotation_name is not None:
        if not isinstance(annotation_name, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"Expected string for annotation_name, got {type(annotation_name)}")
        if not annotation_name.strip():
            raise ValueError("Annotation name cannot be empty or whitespace")

    # Check for name/annotation_name columns
    has_name = "name" in dataframe.columns
    has_annotation_name = "annotation_name" in dataframe.columns
    if has_name and has_annotation_name:
        raise ValueError("DataFrame cannot have both 'name' and 'annotation_name' columns")
    if not annotation_name and not has_name and not has_annotation_name:
        raise ValueError(
            "DataFrame must contain either 'name' or 'annotation_name' column, "
            "or provide a global annotation_name parameter"
        )

    # Check for required columns
    required_columns = set()  # pyright: ignore[reportUnknownVariableType]
    if annotator_kind is None:
        required_columns.add("annotator_kind")  # pyright: ignore[reportUnknownMemberType]

    if not required_columns.issubset(dataframe.columns):
        raise ValueError(
            f"DataFrame must contain columns: {required_columns}. "
            f"Found columns: {dataframe.columns.tolist()}"
        )

    # Check for non-null values in required columns
    for col in required_columns:  # pyright: ignore[reportUnknownVariableType]
        if dataframe[col].isna().all():  # pyright: ignore[reportUnknownMemberType]
            raise ValueError(f"Column '{col}' must contain at least one non-null value")

    # Validate name values if no global name is provided
    if annotation_name is None:
        name_column = "annotation_name" if has_annotation_name else "name"
        # Check for null/NaN values
        if dataframe[name_column].isna().any():  # pyright: ignore[reportUnknownMemberType]
            raise ValueError(f"{name_column} values cannot be None")
        # Check for empty or whitespace-only strings
        if (dataframe[name_column].str.strip() == "").any():  # pyright: ignore[reportUnknownMemberType]
            raise ValueError(f"{name_column} values must be non-empty strings")
        # Check for non-string values
        if not all(isinstance(x, str) for x in dataframe[name_column]):  # pyright: ignore[reportUnknownVariableType,reportUnknownMemberType]
            raise ValueError(f"{name_column} values must be strings")

    # Check for span_id in either columns or index
    has_span_id = "span_id" in dataframe.columns
    has_context_span_id = "context.span_id" in dataframe.columns
    if has_span_id and has_context_span_id:
        raise ValueError("DataFrame cannot have both 'span_id' and 'context.span_id' columns")
    if (
        not has_span_id
        and not has_context_span_id
        and not all(isinstance(x, str) for x in dataframe.index)  # pyright: ignore[reportUnknownVariableType,reportUnknownMemberType]
    ):
        raise ValueError(
            "DataFrame must have either a 'span_id' or 'context.span_id' column, or a string-based index"  # noqa: E501
        )

    # Validate span_id values if using column
    span_id_column = "context.span_id" if has_context_span_id else "span_id"
    if span_id_column in dataframe.columns:
        # Check for None values
        if dataframe[span_id_column].isna().any():  # pyright: ignore[reportUnknownMemberType]
            raise ValueError(f"{span_id_column} values cannot be None")
        # Check for empty or whitespace-only strings
        if (dataframe[span_id_column].str.strip() == "").any():  # pyright: ignore[reportUnknownMemberType]
            raise ValueError(f"{span_id_column} values must be non-empty strings")
        # Check for non-string values
        if not all(isinstance(x, str) for x in dataframe[span_id_column]):  # pyright: ignore[reportUnknownVariableType,reportUnknownMemberType]
            raise ValueError(f"{span_id_column} values must be strings")
    # Validate index values if using index as span_id
    else:
        # Check for empty or whitespace-only strings
        if (pd.Series(dataframe.index).str.strip() == "").any():  # pyright: ignore[reportUnknownMemberType,reportUnknownArgumentType]
            raise ValueError("Index values must be non-empty strings when used as span_id")
        # Check for non-string values
        if not all(isinstance(x, str) for x in dataframe.index):  # pyright: ignore[reportUnknownVariableType,reportUnknownMemberType]
            raise ValueError("Index values must be strings when used as span_id")

    # Check global annotator_kind if provided
    if annotator_kind is not None and annotator_kind not in _VALID_ANNOTATOR_KINDS:
        raise ValueError(
            f"Invalid annotator_kind value: {annotator_kind}. "
            f"Must be one of: {_VALID_ANNOTATOR_KINDS}"
        )

    # Only check row-level annotator_kind values if no global value is provided
    if annotator_kind is None and "annotator_kind" in dataframe.columns:
        invalid_values = set(dataframe["annotator_kind"].dropna().unique()) - _VALID_ANNOTATOR_KINDS  # pyright: ignore[reportUnknownMemberType,reportUnknownArgumentType]
        if invalid_values:
            raise ValueError(
                f"Invalid annotator_kind values found in DataFrame: {invalid_values}. "
                f"Must be one of: {_VALID_ANNOTATOR_KINDS}"
            )


def _chunk_dataframe(
    *,
    dataframe: pd.DataFrame,
    annotation_name: Optional[str] = None,
    annotator_kind: Optional[Literal["LLM", "CODE", "HUMAN"]] = None,
    chunk_size: int = _DATAFRAME_CHUNK_SIZE,
) -> Iterator[list[v1.SpanAnnotationData]]:
    """Internal function to split a DataFrame into smaller chunks for batch processing.

    This function processes the DataFrame in chunks of 100 rows for efficient batch processing.
    It handles type conversion and validation of the data before creating span annotations.

    Args:
        dataframe: The DataFrame to split into chunks. Must contain either a 'span_id' column or have a non-empty index.
        annotation_name: Optional. The name to use for all annotations. If provided, this value will be used
            for all rows and the DataFrame does not need to include a "name" column.
        annotator_kind: Optional. The kind of annotator used for all annotations. If provided, this value will be used
            for all rows and the DataFrame does not need to include an "annotator_kind" column.
            Must be one of "LLM", "CODE", or "HUMAN".

    Yields:
        Lists of SpanAnnotationData objects, one chunk at a time.

    Raises:
        ValueError: If the DataFrame is invalid or if required fields are missing.
        TypeError: If score values cannot be converted to float.
    """  # noqa: E501
    # Validate DataFrame upfront
    _validate_dataframe(
        dataframe=dataframe,
        annotation_name=annotation_name,
        annotator_kind=annotator_kind,
    )

    span_annotations = []
    for idx, row in dataframe.iterrows():  # pyright: ignore[reportUnknownVariableType,reportUnknownMemberType]
        try:
            # Get required fields with null checks
            row_name = annotation_name
            if row_name is None:
                if "name" in row and bool(row["name"]):  # pyright: ignore[reportUnknownArgumentType]
                    row_name = str(row["name"])  # pyright: ignore[reportUnknownArgumentType]
                elif "annotation_name" in row and bool(row["annotation_name"]):  # pyright: ignore[reportUnknownArgumentType]
                    row_name = str(row["annotation_name"])  # pyright: ignore[reportUnknownArgumentType]
            assert row_name
            row_annotator_kind = annotator_kind
            if row_annotator_kind is None:
                row_annotator_kind = cast(
                    _AnnotatorKind,
                    str(row["annotator_kind"])  # pyright: ignore[reportUnknownArgumentType]
                    if "annotator_kind" in row and bool(row["annotator_kind"])  # pyright: ignore[reportUnknownArgumentType]
                    else None,
                )

            # Get span_id from either column or index
            span_id = (
                str(row["span_id"])  # pyright: ignore[reportUnknownArgumentType]
                if "span_id" in dataframe.columns and bool(row["span_id"])  # pyright: ignore[reportUnknownArgumentType]
                else str(row["context.span_id"])  # pyright: ignore[reportUnknownArgumentType]
                if "context.span_id" in dataframe.columns and bool(row["context.span_id"])  # pyright: ignore[reportUnknownArgumentType]
                else str(idx)
            )

            # Get optional fields with proper type conversion
            label = str(row["label"]) if "label" in row and bool(row["label"]) else None  # pyright: ignore[reportUnknownArgumentType]
            score = None
            if "score" in row and row["score"] is not None:
                try:
                    score = float(row["score"])  # pyright: ignore[reportUnknownArgumentType,reportArgumentType]
                except (ValueError, TypeError):
                    raise TypeError(f"Score value '{row['score']}' cannot be converted to float")
            explanation = (
                str(row["explanation"]).strip()  # pyright: ignore[reportUnknownArgumentType]
                if "explanation" in row and bool(row["explanation"])  # pyright: ignore[reportUnknownArgumentType]
                else None
            )
            metadata = cast(
                dict[str, Any],
                dict(row["metadata"]) if "metadata" in row and bool(row["metadata"]) else None,  # pyright: ignore[reportUnknownArgumentType]
            )
            identifier = (
                str(row["identifier"]) if "identifier" in row and bool(row["identifier"]) else None  # pyright: ignore[reportUnknownArgumentType]
            )

            annotation = _get_span_annotation(
                span_id=span_id,
                annotation_name=row_name,  # pyright: ignore[reportArgumentType]
                annotator_kind=row_annotator_kind,  # pyright: ignore[reportArgumentType]
                label=label,
                score=score,
                explanation=explanation,
                metadata=metadata,  # pyright: ignore[reportArgumentType]
                identifier=identifier,
            )

            span_annotations.append(annotation)  # pyright: ignore[reportUnknownMemberType]

            # Yield chunk when we reach chunk_size
            if len(span_annotations) >= chunk_size:  # pyright: ignore[reportUnknownArgumentType]
                yield span_annotations
                span_annotations = []

        except Exception as e:
            raise ValueError(f"Error processing row {idx}: {str(e)}")

    # Yield any remaining annotations
    if span_annotations:
        yield span_annotations



import json
from typing import Any, Iterable, Mapping, Optional

import pytest
from deepdiff.diff import DeepDiff
from faker import Faker
from mistralai.models import (
    AssistantMessageContentTypedDict,
    AssistantMessageTypedDict,
    FunctionCallTypedDict,
    FunctionTypedDict,
    ImageURLChunkTypedDict,
    ImageURLTypedDict,
    SystemMessageContentTypedDict,
    SystemMessageTypedDict,
    TextChunkTypedDict,
    ToolCallTypedDict,
    ToolMessageContentTypedDict,
    ToolMessageTypedDict,
    ToolTypedDict,
    UserMessageContentTypedDict,
    UserMessageTypedDict,
)

from phoenix.client.__generated__.v1 import (
    ImageContentPart,
    PromptMessage,
    TextContentPart,
    TextContentValue,
    ToolCallContentPart,
)
from phoenix.client.helpers.sdk.mistralai.chat_complete import (
    _from_image,
    _from_message,
    _from_text,
    _from_tool_call,
    _from_tools,
    _to_image,
    _to_messages,
    _to_text,
    _to_tool_call,
    _to_tools,
)
from phoenix.client.utils.template_formatters import NO_OP_FORMATTER

fake = Faker()


def _dict() -> dict[str, Any]:
    return fake.pydict(3, value_types=(int, float, bool, str))  # pyright: ignore[reportUnknownMemberType]


def _str() -> str:
    return fake.pystr(8, 8)


def _user_msg(
    content: UserMessageContentTypedDict,
) -> UserMessageTypedDict:
    return UserMessageTypedDict(
        role="user",
        content=content,
    )


def _assistant_msg(
    content: Optional[AssistantMessageContentTypedDict] = None,
    tool_calls: Iterable[ToolCallTypedDict] = (),
) -> AssistantMessageTypedDict:
    if not tool_calls:
        return AssistantMessageTypedDict(
            role="assistant",
            content=content,
        )
    if content is None:
        return AssistantMessageTypedDict(
            role="assistant",
            tool_calls=list(tool_calls),
        )
    return AssistantMessageTypedDict(
        role="assistant",
        content=content,
        tool_calls=list(tool_calls),
    )


def _tool_msg(
    content: ToolMessageContentTypedDict,
) -> ToolMessageTypedDict:
    return ToolMessageTypedDict(
        role="tool",
        content=content,
        tool_call_id=_str(),
    )


def _system_msg(
    content: SystemMessageContentTypedDict,
) -> SystemMessageTypedDict:
    return SystemMessageTypedDict(
        role="system",
        content=content,
    )


def _text() -> TextChunkTypedDict:
    return TextChunkTypedDict(type="text", text=_str())


def _image() -> ImageURLChunkTypedDict:
    return ImageURLChunkTypedDict(
        type="image_url",
        image_url=ImageURLTypedDict(
            url=fake.image_url(),
        ),
    )


def _tool_call() -> ToolCallTypedDict:
    return ToolCallTypedDict(
        id=_str(),
        type="function",
        function=FunctionCallTypedDict(
            name=_str(),
            arguments=json.dumps(_dict()),
        ),
    )


def _tool() -> ToolTypedDict:
    return ToolTypedDict(
        type="function",
        function=FunctionTypedDict(
            name=_str(),
            description=_str(),
            parameters={
                "type": "object",
                "properties": {
                    "x": {"type": "int", "description": _str()},
                    "y": {"type": "string", "description": _str()},
                },
                "required": ["x", "y"],
                "additionalProperties": False,
            },
        ),
    )


class TestUserMessageTypedDict:
    @pytest.mark.parametrize(
        "obj",
        [
            _user_msg(_str()),
            _user_msg([_text(), _text()]),
        ],
    )
    def test_round_trip(self, obj: UserMessageTypedDict) -> None:
        x: PromptMessage = _from_message(obj)
        assert not DeepDiff([obj], list(_to_messages(x, {}, NO_OP_FORMATTER)))


class TestSystemMessageTypedDict:
    @pytest.mark.parametrize(
        "obj",
        [
            _system_msg(_str()),
            _system_msg([_text(), _text()]),
        ],
    )
    def test_round_trip(self, obj: SystemMessageTypedDict) -> None:
        x: PromptMessage = _from_message(obj)
        assert not DeepDiff([obj], list(_to_messages(x, {}, NO_OP_FORMATTER)))


class TestAssistantMessageTypedDict:
    @pytest.mark.parametrize(
        "obj",
        [
            _assistant_msg(_str()),
            _assistant_msg([_text(), _text()]),
            _assistant_msg(None, [_tool_call(), _tool_call()]),
        ],
    )
    def test_round_trip(self, obj: AssistantMessageTypedDict) -> None:
        x: PromptMessage = _from_message(obj)
        assert not DeepDiff([obj], list(_to_messages(x, {}, NO_OP_FORMATTER)))


class TestToolMessageTypedDict:
    @pytest.mark.parametrize(
        "obj",
        [
            _tool_msg(_str()),
            _tool_msg([_text(), _text()]),
        ],
    )
    def test_round_trip(self, obj: ToolMessageTypedDict) -> None:
        x: PromptMessage = _from_message(obj)
        assert not DeepDiff([obj], list(_to_messages(x, {}, NO_OP_FORMATTER)))


class TestToolTypedDict:
    @pytest.mark.parametrize(
        "tools",
        [[_tool() for _ in range(3)]],
    )
    def test_round_trip(self, tools: Iterable[ToolTypedDict]) -> None:
        new_tools = list(_to_tools(_from_tools(tools)))
        assert not DeepDiff(list(tools), new_tools)


class TestTextChunkTypedDict:
    def test_round_trip(self) -> None:
        obj: TextChunkTypedDict = _text()
        x: TextContentPart = _from_text(obj)
        new_obj: TextChunkTypedDict = _to_text(x, {}, NO_OP_FORMATTER)
        assert not DeepDiff(obj, new_obj)

    def test_formatter(self) -> None:
        x = TextContentPart(type="text", text=TextContentValue(text=_str()))
        formatter, variables = _MockFormatter(), _dict()
        ans: TextChunkTypedDict = _to_text(x, variables, formatter)
        assert ans["text"] == formatter.format(x["text"]["text"], variables=variables)


class TestToolCallTypedDict:
    def test_round_trip(self) -> None:
        obj: ToolCallTypedDict = _tool_call()
        x: ToolCallContentPart = _from_tool_call(obj)
        new_obj: ToolCallTypedDict = _to_tool_call(x, {}, NO_OP_FORMATTER)
        assert not DeepDiff(obj, new_obj)


class TestImageURLChunkTypedDict:
    def test_round_trip(self) -> None:
        obj: ImageURLChunkTypedDict = _image()
        x: ImageContentPart = _from_image(obj)
        new_obj: ImageURLChunkTypedDict = _to_image(x, {}, NO_OP_FORMATTER)
        assert not DeepDiff(obj, new_obj)


class _MockFormatter:
    def format(self, _: str, /, *, variables: Mapping[str, str]) -> str:
        return json.dumps(variables)



import functools
import inspect
import random
from abc import ABC
from collections.abc import Awaitable, Mapping, Sequence
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from types import MappingProxyType
from typing import Any, Callable, Optional, Union, cast

from typing_extensions import Protocol, TypedDict, runtime_checkable

from phoenix.client.__generated__ import v1

# Type aliases
JSONSerializable = Optional[Union[dict[str, Any], list[Any], str, int, float, bool]]
ExperimentId = str
DatasetId = str
DatasetVersionId = str
ExampleId = str
RepetitionNumber = int
ExperimentRunId = str
TraceId = str
TaskOutput = JSONSerializable
ExampleOutput = Mapping[str, JSONSerializable]
ExampleMetadata = Mapping[str, JSONSerializable]
ExampleInput = Mapping[str, JSONSerializable]
Score = Optional[Union[bool, int, float]]
Label = Optional[str]
Explanation = Optional[str]
EvaluatorName = str
EvaluatorKind = str

DRY_RUN = "DRY_RUN"

# Use autogenerated types
Experiment = v1.Experiment
ExperimentRun = v1.ExperimentRunResponse
EvaluationResult = v1.ExperimentEvaluationResult


class AnnotatorKind(Enum):
    CODE = "CODE"
    LLM = "LLM"


def _dry_run_id() -> str:
    suffix = random.getrandbits(24).to_bytes(3, "big").hex()
    return f"{DRY_RUN}_{suffix}"


@dataclass(frozen=True)
class TestCase:
    example: v1.DatasetExample
    repetition_number: RepetitionNumber


@dataclass(frozen=True)
class ExperimentEvaluationRun:
    experiment_run_id: ExperimentRunId
    start_time: datetime
    end_time: datetime
    name: str
    annotator_kind: str
    error: Optional[str] = None
    result: Optional[EvaluationResult] = None
    id: str = field(default_factory=_dry_run_id)
    trace_id: Optional[TraceId] = None
    metadata: Mapping[str, JSONSerializable] = field(default_factory=dict)

    def __post_init__(self) -> None:
        if self.result is None and self.error is None:
            raise ValueError("Must specify either result or error")


# Task and Evaluator types
ExperimentTask = Union[
    Callable[[v1.DatasetExample], TaskOutput],
    Callable[[v1.DatasetExample], Awaitable[TaskOutput]],
    Callable[..., JSONSerializable],
    Callable[..., Awaitable[JSONSerializable]],
]

EvaluatorOutput = Union[
    EvaluationResult, bool, int, float, str, tuple[Score, Label, Explanation], dict[str, Any]
]


@runtime_checkable
class Evaluator(Protocol):
    """
    Protocol for evaluators that can score experiment outputs.

    Any object implementing this protocol can be used as an evaluator.
    Subclasses should implement either the `evaluate` or `async_evaluate` method.
    Implementing both methods is recommended, but not required.
    """

    @property
    def name(self) -> str:
        """The name of the evaluator."""
        ...

    @property
    def kind(self) -> str:
        """The kind of evaluator (e.g., 'CODE', 'LLM')."""
        ...

    def evaluate(
        self,
        *,
        output: Optional[TaskOutput] = None,
        expected: Optional[ExampleOutput] = None,
        metadata: ExampleMetadata = MappingProxyType({}),
        input: ExampleInput = MappingProxyType({}),
        **kwargs: Any,
    ) -> EvaluationResult:
        """Evaluate the output synchronously."""
        ...

    async def async_evaluate(
        self,
        *,
        output: Optional[TaskOutput] = None,
        expected: Optional[ExampleOutput] = None,
        metadata: ExampleMetadata = MappingProxyType({}),
        input: ExampleInput = MappingProxyType({}),
        **kwargs: Any,
    ) -> EvaluationResult:
        """Evaluate the output asynchronously."""
        ...


def _validate_evaluator_signature(sig: inspect.Signature) -> None:
    """Validate that a function signature is compatible with evaluator requirements."""
    params = sig.parameters
    valid_named_params = {"input", "output", "expected", "reference", "metadata"}
    if len(params) == 0:
        raise ValueError("Evaluator function must have at least one parameter.")
    if len(params) > 1:
        for param_name in set(params) - valid_named_params:
            param = params[param_name]
            if (
                param.kind is inspect.Parameter.VAR_KEYWORD
                or param.default is not inspect.Parameter.empty
            ):
                continue
            raise ValueError(
                f"Invalid parameter name in evaluator function: {param_name}. "
                "Parameter names for multi-argument functions must be "
                f"any of: {', '.join(valid_named_params)}."
            )


def _validate_evaluator_method_signature(fn: Callable[..., Any], fn_name: str) -> None:
    """Validate that an evaluator method has the correct signature."""
    sig = inspect.signature(fn)
    _validate_evaluator_signature(sig)
    for param in sig.parameters.values():
        if param.kind is inspect.Parameter.VAR_KEYWORD:
            return
    else:
        raise ValueError(f"`{fn_name}` should allow variadic keyword arguments `**kwargs`")


class BaseEvaluator(ABC):
    """
    A helper abstract class to guide the implementation of an `Evaluator` object.
    Subclasses must implement either the `evaluate` or `async_evaluate` method.
    Implementing both methods is recommended, but not required.

    This Class is intended to be subclassed, and should not be instantiated directly.
    """

    _kind: AnnotatorKind
    _name: EvaluatorName

    @property
    def name(self) -> EvaluatorName:
        if hasattr(self, "_name"):
            return self._name
        return self.__class__.__name__

    @property
    def kind(self) -> EvaluatorKind:
        if hasattr(self, "_kind"):
            return self._kind.value
        return AnnotatorKind.CODE.value

    def __new__(cls, *args: Any, **kwargs: Any) -> "BaseEvaluator":
        if cls is BaseEvaluator:
            raise TypeError(f"{cls.__name__} is an abstract class and should not be instantiated.")
        return object.__new__(cls)

    def evaluate(
        self,
        *,
        output: Optional[TaskOutput] = None,
        expected: Optional[ExampleOutput] = None,
        metadata: ExampleMetadata = MappingProxyType({}),
        input: ExampleInput = MappingProxyType({}),
        **kwargs: Any,
    ) -> EvaluationResult:
        """
        Evaluate the output synchronously.

        For subclassing, one should implement either this sync method or the
        async version. Implementing both is recommended but not required.
        """
        raise NotImplementedError

    async def async_evaluate(
        self,
        *,
        output: Optional[TaskOutput] = None,
        expected: Optional[ExampleOutput] = None,
        metadata: ExampleMetadata = MappingProxyType({}),
        input: ExampleInput = MappingProxyType({}),
        **kwargs: Any,
    ) -> EvaluationResult:
        """
        Evaluate the output asynchronously.

        For subclassing, one should implement either this async method or the
        sync version. Implementing both is recommended but not required.
        """
        return self.evaluate(
            output=output,
            expected=expected,
            metadata=metadata,
            input=input,
            **kwargs,
        )

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        # Skip validation for abstract classes
        if getattr(cls, "__abstract__", False):
            return

        # Validate that subclass implements at least one evaluation method
        evaluate_fn_signature = inspect.signature(BaseEvaluator.evaluate)
        for super_cls in inspect.getmro(cls):
            if super_cls is BaseEvaluator:
                break
            if evaluate := super_cls.__dict__.get(BaseEvaluator.evaluate.__name__):  # pyright: ignore[reportUnknownVariableType]
                # pyright: ignore[reportUnknownVariableType] - suppress type warnings for evaluate variable
                if isinstance(evaluate, classmethod):
                    evaluate = evaluate.__func__  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                evaluate = cast(Callable[..., Any], evaluate)  # pyright: ignore[reportUnknownVariableType] - Cast to fix pyright warnings
                assert callable(evaluate), "`evaluate()` method should be callable"
                # need to remove the first param, i.e. `self`
                _validate_evaluator_method_signature(functools.partial(evaluate, None), "evaluate")  # pyright: ignore[reportUnknownArgumentType,reportUnknownVariableType]
                return
            if async_evaluate := super_cls.__dict__.get(BaseEvaluator.async_evaluate.__name__):  # pyright: ignore[reportUnknownVariableType]
                if isinstance(async_evaluate, classmethod):
                    async_evaluate = async_evaluate.__func__  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                async_evaluate = cast(
                    Callable[..., Any], async_evaluate
                )  # Cast to fix pyright warnings
                assert callable(async_evaluate), "`async_evaluate()` method should be callable"
                # need to remove the first param, i.e. `self`
                _validate_evaluator_method_signature(
                    functools.partial(async_evaluate, None),
                    "async_evaluate",  # pyright: ignore[reportUnknownArgumentType,reportUnknownVariableType]
                )
                return

        raise ValueError(
            f"Evaluator must implement either "
            f"`def evaluate{evaluate_fn_signature}` or "
            f"`async def async_evaluate{evaluate_fn_signature}`"
        )


class RanExperiment(TypedDict):
    """
    A completed experiment with its results.

    This represents an experiment that has been run and contains both the experiment
    metadata and the task runs. It can be used as input to evaluate_experiment to
    add additional evaluations.
    """

    experiment_id: ExperimentId
    dataset_id: DatasetId
    task_runs: list[ExperimentRun]
    evaluation_runs: list[ExperimentEvaluationRun]
    experiment_metadata: Mapping[str, Any]


# Type aliases for evaluators
ExperimentEvaluator = Union[
    Evaluator,
    Callable[..., EvaluatorOutput],
    Callable[..., Awaitable[EvaluatorOutput]],
]
ExperimentEvaluators = Union[
    ExperimentEvaluator,
    Sequence[ExperimentEvaluator],
    Mapping[EvaluatorName, ExperimentEvaluator],
]
Evaluators = Union[
    ExperimentEvaluator,
    Sequence[ExperimentEvaluator],
    Mapping[EvaluatorName, ExperimentEvaluator],
]
RateLimitErrors = Union[type[BaseException], Sequence[type[BaseException]]]
